"""Intra-component communication support.

Message types and protocol support for control signals,
queue management, and work dispatching.
"""
__all__ = ("Command", "QueueItem", "CommandQueueControlItem", "CommandQueueAddItem", "StopCommand")

import logging
import typing
import weakref
from abc import abstractmethod

from scalems.exceptions import APIError

logger = logging.getLogger(__name__)
logger.debug("Importing {}".format(__name__))


class QueueItem(dict, typing.MutableMapping[str, typing.Union[str, bytes]]):
    """Queue items are either workflow items or control messages.

    Control messages are indicated by the key ``'command'``, and are described
    by :py:class:`CommandQueueControlItem`.

    Workflow items are indicated by the key ``'add_item'``, and are described by
    :py:class:`CommandQueueAddItem`.
    """

    def _hexify(self):
        """Allow binary fields to be printable."""
        for key, value in self.items():
            if isinstance(value, bytes):
                value = value.hex()
            yield key, value

    def __str__(self) -> str:
        return str(dict(self._hexify()))


class CommandQueueControlItem(QueueItem, typing.MutableMapping[str, str]):
    """String-encoded Command for the Executor command queue.

    Instructions for the `RuntimeManager`, intercepted and processed by
    :py:func:`scalems.execution.manage_execution()`.

    Currently, the only supported key is "command".

    Supported commands may grow to comprise a Compute Provide Interface.
    """

    _allowed: typing.ClassVar = {"command": {"hello", "stop", "version"}}

    def __setitem__(self, k: str, v: str) -> None:
        if k in self._allowed:
            if v in self._allowed[k]:
                super().__setitem__(k, v)
            else:
                raise APIError(f"Unsupported command value: {repr(v)}")
        else:
            raise APIError(f"Unsupported command key: {repr(k)}")


class CommandQueueAddItem(QueueItem, typing.MutableMapping[str, bytes]):
    """String-encoded add_item command for the Executor command queue.

    The intended payload is an item to be added to the workflow graph:
    e.g. an operation, data reference, subgraph, or something meaningful to an
    :py:class:`~scalems.execution.AbstractWorkflowUpdater`.

    In practice, the value may just be a token or identifier. In the initial
    version of the scalems protocol, the value is passed to the item editor factory
    that is obtained from :py:func:`scalems.execution.RuntimeManager.get_edit_item()`.
    """

    _allowed: typing.ClassVar = {"add_item"}

    def __setitem__(self, k: str, v: bytes) -> None:
        if k in self._allowed:
            if isinstance(v, bytes):
                super().__setitem__(k, v)
            else:
                raise APIError(f"Unsupported add_item key: {repr(v)}")
        else:
            raise APIError(f"Unsupported command: {repr(k)}")


SerializedValueT = typing.TypeVar("SerializedValueT", bound=typing.Union[str, bytes])


class Command(typing.Protocol[SerializedValueT]):
    """An instruction to the workflow execution run time machinery.

    A limited schema for Command objects allows abstraction of the workflow
    management and run time collaborations to be represented in a way that is
    wire-protocol-friendly, easily serializable, and easily converted between
    structured data types or file formats.

    Commands of different types (represented by subclasses of :py:class:`~scalems.messages.Command`)
    are directed to different components of the SCALE-MS workflow management model.

    Support for serialization follows the model of the object encoding/decoding
    hooks in the :py:mod:`json` module.

    Immediate realizations of the Command Protocol are Abstract Base Classes
    that map themselves to "command category" strings, which serve as keys in
    the serialized mapping object representation.

    The Abstract Base Classes are responsible for translating the serialized
    object value into a Command subclass instance.
    """

    key: typing.ClassVar[str]
    """Key word for mapping a named command to the Command creation function."""

    # TODO(#23): This class hierarchy needs rethinking.
    # This is a reasonable attribute of an Abstract Base Class,
    # but should not be part of a Protocol definition since it is not a behavior
    # that we want to specify for all instances.
    __subtype: typing.ClassVar[
        typing.MutableMapping[str, typing.Callable[..., typing.Callable[[SerializedValueT], typing.Any]]]
    ] = {}
    """Map named command categories to WeakMethod weakrefs to callable decoders."""

    @classmethod
    @abstractmethod
    def create(cls, command: SerializedValueT) -> "Command":
        """Creation function will be registered during subclassing.

        The *command* argument will be provided with the *value*
        of the deserialized Mapping entry during *decode()*.
        """
        raise NotImplementedError

    @staticmethod
    def decode(obj: typing.Mapping[str, SerializedValueT]) -> "Command":
        """Convert a deserialized object representation to a Command instance.

        Abstract Command base classes
        """
        items = tuple(obj.items())
        if len(items) != 1 or len(items[0]) != 2:
            raise ValueError("Not a key-value pair. decode() expects a (single-element) Mapping.")
        key, value = items[0]
        creation_function = Command.__subtype[key]()
        return creation_function(value)

    @abstractmethod
    def encode(self):
        """Produce a serializeable object record."""
        # Example:
        #   return {self.key: self.value}
        raise NotImplementedError

    def __init_subclass__(cls, **kwargs):
        """Register an abstract command base class for dispatched creation."""
        # Check the protocol.
        if cls is not Command:
            assert hasattr(cls, "key")
            assert cls.key not in Command.__subtype
            Command.__subtype[cls.key] = weakref.WeakMethod(cls.create)
            super().__init_subclass__(**kwargs)


class Control(Command[str], typing.Protocol):
    """Represent a "control" message.

    Supported commands for runtime components.

    Abstract Base Class for Control commands.
    Serialized Control objects are identified by the key word "control".
    The payload (or serialized object) is just a string naming the control command.

    Note: I'm not sure that it is appropriate for concrete Control types to be
    subclasses of Control, and, thus, Command. It is convenient metaprogramming,
    but what we want is more akin to the `register` interaction of explicitly _virtual_
    subclasses, or of the dispatching function objects from functools. I'm withholding
    judgement or redesign until seeing how things play out with command types that
    take arguments, or more complete functionality like dispatchable tasks. At that
    point, though, I expect a decorator pattern will make more sense.
    """

    key: typing.ClassVar = "control"

    message: typing.ClassVar[str]

    # TODO(#23): This class hierarchy needs rethinking.
    # This is a reasonable attribute of an Abstract Base Class,
    # but should not be part of a Protocol definition since it is not a behavior
    # that we want to specify for all instances.
    __control_type: typing.ClassVar[typing.MutableMapping[str, type]] = weakref.WeakValueDictionary()
    """Map named control commands to their concrete types."""

    def __init_subclass__(cls, **kwargs):
        assert hasattr(cls, "message")
        assert cls.message not in Control.__control_type
        Control.__control_type[cls.message] = cls
        # The Control abstract base class has separate semi-private relationships
        # with its parents and with its children. Control insulates subclasses
        # from other base classes, and we interrupt the init_subclass protocol
        # here.
        # Do not call super().__init_subclass__(**kwargs)

    @classmethod
    def create(cls, command: str) -> "Command":
        assert cls is Control
        return cls.__control_type[command]()

    def encode(self):
        return {self.key: self.message}


class StopCommand(Control):
    message: typing.ClassVar[str] = "stop"
    """Announce the end of the session, directing components to shut down cleanly.

    No further commands will be processed after receiving this signal.
    """

    def __init__(self):
        # With a Protocol parent, we need an __init__ to establish this as a concrete class.
        ...


class HelloCommand(Control):
    message: typing.ClassVar[str] = "hello"
    """Elicit a response.

    The structure of the response is a detail of the backend, but probably
    includes version and configuration information to support compatibility
    checks.
    """

    def __init__(self):
        # With a Protocol parent, we need an __init__ to establish this as a concrete class.
        ...


class AddItem(Command[str]):
    """Announce an item to be added to the target workflow.

    Direct an update to the workflow record, such as to describe a data source,
    register a resource, or add a task.

    The payload is a json-serialized ScalemsRaptorWorkItem. An instance can
    be provided as an initializer argument for a backend specific CpiAddItem.

    Design note:
        WorkflowManager.add_item() and CommandQueueAddItem are not normalized or
        normative. We can start engineering scalems.messages.AddItem
        and scalems.radical.raptor.CpiAddItem together from the bottom up and then
        make another pass at the higher level / entry point interface.

    """

    key: typing.ClassVar = "add_item"

    def __init__(self, encoded_item: str):
        # TODO: This should be a public scalems schema, but is currently assumed
        #  to be scalems.radical.raptor.ScalemsRaptorWorkItem.
        self._encoded_item = encoded_item

    @classmethod
    def create(cls, command: str) -> "Command":
        return cls(command)

    def encode(self):
        return {self.key: self._encoded_item}

    @property
    def encoded_item(self):
        return self._encoded_item



"""Runtime management for the scalems client.

Provide the framework for the SCALE-MS execution middleware layer.

Execution dispatching implementations can be chosen from the command line when
the interpreter is launched with Python's ``-m`` command line option.

Alternatively, the script can use a Python Context Manager (a ``with`` block) to
activate a WorkflowContext with an execution dispatcher implementation.

It is inadvisable to activate an execution dispatching environment with
procedural calls because it is more difficult to make sure that acquired resources
are properly released, in case the interpreter has to terminate prematurely.

Execution dispatching generally uses some sort of concurrency model,
but there is not a consistent concurrency model for all dispatchers.
ScaleMS provides abstractions to insulate scripting from particular implementations
or concurrency primitives (such as the need to call :py:func:`asyncio.run`).

The following diagram uses the :py:mod:`scalems.radical` execution module
to illustrate the workflow execution.

.. uml::

    title scalems on radical.pilot (client side)

    participant "workflow script" as script
    box "SCALE-MS framework" #honeydew
    participant "SCALE-MS API" as scalems.Runtime
    participant WorkflowManager as client_workflowmanager
    participant Queuer
    end box
    box "SCALE-MS RP adapter" #linen
    participant scalems.radical
    participant Executor as client_executor
    end box

    autoactivate on

    -> scalems.radical: python -m scalems.radical ...

    scalems.radical -> scalems.Runtime: scalems.invocation.run(workflow_manager)
    scalems.Runtime -> scalems.radical: configuration()
    return
    scalems.Runtime -> scalems.radical: workflow_manager()
    scalems.radical -> client_workflowmanager **: <<create>>
    scalems.Runtime <-- scalems.radical:

    scalems.Runtime -> script: <<runpy>>
    return @scalems.app

    scalems.Runtime -> scalems.Runtime: run_dispatch(work, context)
    scalems.Runtime -> client_workflowmanager: async with dispatch()

    client_workflowmanager -> client_workflowmanager: executor_factory()

    client_workflowmanager -> client_executor **: <<create>>
    client_workflowmanager --> client_workflowmanager: executor
    client_workflowmanager -> Queuer **: <<create>>

    == Launch runtime ==

    client_workflowmanager -> client_executor: async with executor
    activate client_workflowmanager #lightgray

    ...Dispatch work. See `manage_execution`...

    client_workflowmanager <-- client_executor: leave executor context
    deactivate client_workflowmanager
    scalems.Runtime <-- client_workflowmanager: end dispatching context
    scalems.Runtime --> scalems.Runtime: loop.run_until_complete()
    deactivate client_workflowmanager
    scalems.Runtime --> scalems.radical: SystemExit.code
    deactivate scalems.Runtime
    <-- scalems.radical: sys.exit

The interface available to ``@scalems.app`` is under development.
See :py:mod:`scalems.workflow`.

The details of work dispatching are not yet strongly specified or fully encapsulated.
`manage_execution` mediates a collaboration between a `RuntimeManager` and a
`WorkflowManager` (via `AbstractWorkflowUpdater`).

"""

from __future__ import annotations

__all__ = ("AbstractWorkflowUpdater", "RuntimeManager", "manage_execution")

import abc
import asyncio
import contextlib
import logging
import typing

import scalems.exceptions
from scalems.exceptions import APIError
from scalems.exceptions import DispatchError
from scalems.exceptions import InternalError
from scalems.exceptions import MissingImplementationError
from scalems.exceptions import ProtocolError
from scalems.messages import QueueItem
from scalems.store import FileStore
from scalems.workflow import Task

logger = logging.getLogger(__name__)
logger.debug("Importing {}".format(__name__))


class AbstractWorkflowUpdater(abc.ABC):
    # TODO: The `item: Task` argument to `AbstractWorkflowUpdater.submit()` should be
    #    decoupled from the WorkflowManager implementation.
    @abc.abstractmethod
    async def submit(self, *, item: Task) -> asyncio.Task:
        """Submit work to update a workflow item.

        Args:
            item: managed workflow item as produced by
                  WorkflowManager.edit_item()

        Returns:
            watcher task for a coroutine that has already started.

        The returned Task is a watcher task to confirm completion of or allow
        cancellation of the dispatched work. Additional hidden Tasks may be
        responsible for updating the provided item when results are available,
        so the result of this task is not necessarily directly interesting to
        the execution manager.

        The caller check for unexpected cancellation or early failure. If the task has
        not failed or been canceled by the time *submit()* returns, then we can
        conclude that the task has been appropriately bound to the workfow item.
        """
        ...


class RuntimeDescriptor:
    """Data Descriptor class for (backend-specific) runtime state access.

    TODO: Consider making this class generic in terms of the backend/configuration
     type. (Maybe only possible if we keep the subclassing model of RuntimeManagers rather
     than composing type information at instantiation.)
    """

    # Ref: https://docs.python.org/3/reference/datamodel.html#implementing-descriptors
    def __set_name__(self, owner, name):
        # Called by type.__new__ during class creation to allow customization.
        if getattr(owner, self.private_name, None) is not None:
            raise ProtocolError(
                f"Cannot assign {repr(self)} to {owner.__qualname__} "
                f"with an existing non-None {self.private_name} member."
            )

    def __get__(self, instance, owner):
        # Note that instance==None when called through the *owner* (as a class attribute).
        if instance is None:
            return self
        else:
            return getattr(instance, self.private_name, None)

    def __set__(self, instance, value):
        if getattr(instance, self.private_name, None) is not None:
            raise APIError("Cannot overwrite an existing runtime state.")
        setattr(instance, self.private_name, value)

    def __delete__(self, instance):
        try:
            delattr(instance, self.private_name)
        except AttributeError:
            pass

    def __init__(self):
        self.private_name = "_runtime_state"


_BackendT = typing.TypeVar("_BackendT")


class RuntimeManager(typing.Generic[_BackendT], abc.ABC):
    """Client side manager for dispatching work loads and managing data flow.

    A RuntimeManager is instantiated within the scope of the
    `scalems.workflow.WorkflowManager.dispatch` context manager using the
    :py:meth:`scalems.workflow.WorkflowManager._executor_factory`.
    """

    get_edit_item: typing.Callable[[], typing.Callable]
    """Get the function that creates a WorkflowItem editing context."""

    datastore: FileStore
    submitted_tasks: typing.MutableSet[asyncio.Task]

    _runtime_configuration: _BackendT

    _command_queue: asyncio.Queue
    _dispatcher_lock: asyncio.Lock
    _queue_runner_task: typing.Union[None, asyncio.Task] = None

    runtime = RuntimeDescriptor()
    """Get/set the current runtime state information.

    Attempting to overwrite an existing *runtime* state raises an APIError.

    De-initialize the stored runtime state by calling ``del`` on the attribute.

    To re-initialize, de-initialize and then re-assign.

    Design Note:
        This pattern allows runtime state objects to be removed from the instance
        data while potentially still in use (such as during clean-up), allowing
        clearer scoping of access to the runtime state object.

        In the future, we may find it more practical to use a module-level "ContextVar"
        and to manage the scope in terms of PEP-567 Contexts. For the initial
        implementation, though, we are using module-level runtime configuration
        information, but storing the state information for initialized runtime facilities
        through this Data Descriptor on the RuntimeManager instance.

    Raises:
        APIError if *state* is not None and a runtime state already exists.

    """

    def __init__(
        self,
        *,
        editor_factory: typing.Callable[[], typing.Callable] = None,
        datastore: FileStore = None,
        loop: asyncio.AbstractEventLoop,
        configuration: _BackendT,
        dispatcher_lock=None,
    ):
        self.submitted_tasks = set()

        # TODO: Only hold a queue in an active context manager.
        self._command_queue = asyncio.Queue()
        self._exception = None
        self._loop: asyncio.AbstractEventLoop = loop

        if editor_factory is None or not callable(editor_factory):
            raise TypeError("Provide a callable that produces an edit_item interface.")
        self.get_edit_item = editor_factory

        if datastore is None:
            raise TypeError("Provide a datastore.")
        self.datastore = datastore

        # TODO: Consider relying on module ContextVars and contextvars.Context scope.
        self._runtime_configuration = configuration

        if not isinstance(dispatcher_lock, asyncio.Lock):
            raise TypeError("An asyncio.Lock is required to control dispatcher state.")
        self._dispatcher_lock = dispatcher_lock

    def configuration(self) -> _BackendT:
        return self._runtime_configuration

    @staticmethod
    async def cpi(command: str, runtime):
        """Dispatcher for CPI messages.

        TODO: Return value? We probably want to be able to capture something we can
            query for the result of the CPI message.
        """
        logger.debug(f"Null CPI handler received command {command}.")

    @staticmethod
    def runtime_shutdown(runtime):
        """Shutdown hook for runtime facilities.

        Called while exiting the context manager. Allows specialized handling of
        runtime backends in terms of the RuntimeManager instance and an abstract
        *Runtime* object, presumably acquired while entering the context manager
        through the *runtime_startup* hook.

        The method is static as a reminder that all state should come through
        the *runtime* argument. The hook is used in the context manager
        implemented by the base class, which manages the removal of state
        information regardless of the actions (or errors) of subclasses using
        this hook.
        """
        pass

    def queue(self):
        # TODO: Only expose queue while in an active context manager.
        return self._command_queue

    def exception(self) -> typing.Union[None, Exception]:
        return self._exception

    @contextlib.contextmanager
    def runtime_configuration(self):
        """Runtime startup context manager hook.

        This allows arbitrary environment changes during the scope of the __aenter__()
        function, as well as a chance to modify the stored runtime configuration object.

        Warning: this is probably not a permanent part of the interface. We should settle
        on whether to store a configuration/context object or rely on the current
        contextvars.Context.
        """
        ...
        try:
            yield None
        finally:
            ...

    @abc.abstractmethod
    async def runtime_startup(self) -> asyncio.Task:
        """Runtime startup hook.

        If the runtime manager uses a Session, this is the place to acquire it.

        This coroutine itself returns a `asyncio.Task`. This allows the caller to yield until
        until the runtime manager is actually ready.

        Implementations may perform additional checks before returning to ensure that
        the runtime queue processing task has run far enough to be well behaved if it
        later encounters an error or is subsequently canceled. Also, blocking
        interfaces used by the runtime manager could be wrapped in separate thread
        executors so that the asyncio event loop doesn't have to block on this call.
        """
        ...

    @abc.abstractmethod
    def updater(self) -> AbstractWorkflowUpdater:
        """Initialize a WorkflowUpdater for the configured runtime."""
        # TODO: Convert from an abstract method to a registration pattern.
        ...

    async def __aenter__(self):
        with self.runtime_configuration():
            try:
                # Get a lock while the state is changing.
                # Warning: The dispatching protocol is immature.
                # Initially, we don't expect contention for the lock, and if there is
                # contention, it probably represents an unintended race condition
                # or systematic dead-lock.
                # TODO: Clarify dispatcher state machine and remove/replace assertions.
                assert not self._dispatcher_lock.locked()
                async with self._dispatcher_lock:
                    runner_task: asyncio.Task = await self.runtime_startup()
                    if runner_task.done():
                        if runner_task.cancelled():
                            raise DispatchError("Runner unexpectedly canceled while starting dispatching.")
                        else:
                            e = runner_task.exception()
                            if e:
                                logger.exception("Runner task failed with an exception.", exc_info=e)
                                raise e
                            else:
                                logger.warning("Runner task stopped unusually early, but did not raise an exception.")
                    self._queue_runner_task = runner_task

                # Note: it would probably be most useful to return something with a
                # WorkflowManager interface...
                return self
            except Exception as e:
                self._exception = e
                raise e

    async def __aexit__(self, exc_type, exc_val, exc_tb):  # noqa: C901
        """Clean up at context exit.

        In addition to handling exceptions, clean up any Session resource.

        We also need to make sure that we properly disengage from any queues
        or generators.

        We can also leave flags for ourself to be checked at __await__,
        if there is a Task associated with the Executor.
        """
        # Note that this coroutine could take a long time and could be
        # cancelled at several points.
        cancelled_error = None
        # The dispatching protocol is immature.
        # Initially, we don't expect contention for the lock, and if there is
        # contention, it probably represents an unintended race condition
        # or systematic dead-lock.
        # TODO: Clarify dispatcher state machine and remove/replace assertions.
        assert not self._dispatcher_lock.locked()
        async with self._dispatcher_lock:
            runtime = self.runtime
            # This method is not thread safe, but we try to make clear as early as
            # possible that instance.session is no longer publicly available.
            del self.runtime

            try:
                # Stop the executor.
                logger.debug("Stopping workflow execution.")
                # TODO: Make sure that nothing else will be adding items to the
                #  queue from this point.
                # We should establish some assurance that the next line represents
                # the last thing that we will put in the queue.
                self._command_queue.put_nowait({"control": "stop"})
                # TODO: Consider what to do differently when we want to cancel work
                #  rather than just finalize it.

                # Currently, the queue runner does not place subtasks,
                # so there is only one thing to await.
                # TODO: We should probably allow the user to provide some sort of timeout,
                #  or infer one from other time limits.
                try:
                    await self._queue_runner_task
                except asyncio.CancelledError as e:
                    raise e
                except Exception as queue_runner_exception:
                    logger.exception("Unhandled exception when stopping queue handler.", exc_info=True)
                    self._exception = queue_runner_exception
                else:
                    logger.debug("Queue runner task completed.")
                finally:
                    if not self._command_queue.empty():
                        logger.error("Command queue never emptied.")

                    if len(self.submitted_tasks) == 0:
                        logger.debug("No tasks to wait for. Continuing to shut down.")
                    else:
                        # Wait for the tasks.
                        # For scalems.radical, the returned values are the rp.Task
                        # objects in their final states.
                        # QUESTION: How long should we wait before canceling tasks?
                        # TODO: Handle some sort of job maximum wall time parameter.
                        results = await asyncio.gather(*self.submitted_tasks)
                        # TODO: Log something useful about the results.
                        assert len(results) == len(self.submitted_tasks)
            except asyncio.CancelledError as e:
                logger.debug(f"{self.__class__.__qualname__} context manager received cancellation while exiting.")
                cancelled_error = e
            except Exception as e:
                logger.exception("Exception while stopping dispatcher.", exc_info=True)
                if self._exception:
                    logger.error("Queuer is already holding an exception.")
                else:
                    self._exception = e
            finally:
                try:
                    await asyncio.to_thread(self.runtime_shutdown, runtime)
                except asyncio.CancelledError as e:
                    cancelled_error = e
                except Exception as e:
                    logger.exception(f"Exception while shutting down {repr(runtime)}.", exc_info=e)
                else:
                    logger.debug("Runtime resources closed.")
        if cancelled_error:
            raise cancelled_error

        # Only return true if an exception should be suppressed (because it was handled).
        # TODO: Catch internal exceptions for useful logging and user-friendliness.
        if exc_type is not None:
            return False


async def manage_execution(executor: RuntimeManager, *, processing_state: asyncio.Event):
    """Process workflow messages until a stop message is received.

    Initial implementation processes commands serially without regard for possible
    concurrency.

    Towards concurrency:
        We can create all tasks without awaiting any of them.

        Some tasks will be awaiting results from other tasks.

        All tasks will be awaiting a `asyncio.Lock` or `asyncio.Condition` for each
        required resource, but must do so indirectly.

        To avoid dead-locks, we can't have a Lock object for each resource unless
        they are managed by an intermediary that can do some serialization of requests.
        In other words, we need a Scheduler that tracks the resource pool,
        packages resource locks only when they can all be acquired without race
        conditions or blocking,
        and which then notifies the Condition for each task that it is allowed to run.

        It should not do so until the dependencies of the task are known to have
        all of the resources they need to complete (running with any dynamic dependencies
        also running) and, preferably, complete.

        Alternatively, the Scheduler can operate in blocks, allocating all resources,
        offering the locks to tasks, waiting for all resources to be released,
        then repeating.
        We can allow some conditions to "wake up" the scheduler to back fill a block
        of resources, but we should be careful with that.

        (We still need to consider dynamic tasks that
        generate other tasks. I think the only way to distinguish tasks which can't be
        dynamic from those which might be would be with the `def` versus `async def` in
        the implementing function declaration. If we abstract `await` with ``scalems.wait``,
        we can throw an exception at execution time after checking a ContextVar.
        It may be better to just let implementers use `await` for dynamically created
        tasks, but we need to make the same check if a function calls ``.result()`` or
        otherwise tries to create a dependency on an item that was not allocated
        resources before the function started executing.
        In a conservative first draft, we can simply throw an exception if
        a non-`async def` function attempts to call a scalems workflow command like
        "add_item" while in an executing context.)

    """
    queue = executor.queue()
    updater = executor.updater()
    # Get a reference to the runtime, since it is removed from the executor before
    # RuntimeManager.__aexit__() shuts down this task.
    runtime = executor.runtime

    # Acknowledge that the coroutine is running and will immediately begin processing
    # queue items.
    processing_state.set()
    # Note that if an exception is raised here, the queue will never be processed.

    # Could also accept a "stop" Event object for the loop conditional,
    # but we would still need a way to yield on an empty queue until either
    # the "stop" event or an item arrives, and then we might have to account
    # for queues that potentially never yield any items, such as by sleeping briefly.
    # We should be able to do
    #     signal_task = asyncio.create_task(stop_event.wait())
    #     queue_getter = asyncio.create_task(command_queue.get())
    #     waiter = asyncio.create_task(asyncio.wait((signal_task, queue_getter),
    #     return_when=FIRST_COMPLETED))
    #     while await waiter:
    #         done, pending = waiter.result()
    #         assert len(done) == 1
    #         if signal_task in done:
    #             break
    #         else:
    #             command: QueueItem = queue_getter.result()
    #         ...
    #         queue_getter = asyncio.create_task(command_queue.get())
    #         waiter = asyncio.create_task(asyncio(wait((signal_task, queue_getter),
    #         return_when=FIRST_COMPLETED))
    #
    command: QueueItem
    while command := await queue.get():
        # Developer note: The preceding line and the following try/finally block are
        # coupled!
        # Once we have awaited asyncio.Queue.get(), we _must_ have a corresponding
        # asyncio.Queue.task_done(). For tidiness, we immediately enter a `try` block
        # with a `finally` suite. Don't separate these without considering the Queue
        # protocol.

        try:
            if not len(command.items()) == 1:
                raise ProtocolError("Expected a single key-value pair.")
            logger.debug(f"Processing command {repr(command)}")

            # TODO(#23): Use formal RPC protocol.
            if "control" in command:
                logger.debug(f"Execution manager received {command['control']} command for {runtime}.")
                try:
                    await executor.cpi(command["control"], runtime)
                except scalems.exceptions.ScopeError as e:
                    logger.debug(
                        f"Control command \"{command['control']}\" ignored due to inactive RuntimeManager.", exc_info=e
                    )
                    # We should probably break here, too, right?
                    # TODO: Handle command results and take appropriate action on failures.
                if command["control"] == "stop":
                    # End queue processing.
                    break
            if "add_item" not in command:
                # This will end queue processing. Tasks already submitted may still
                # complete. Tasks subsequently added will update the static part of
                # the workflow (WorkflowManager) and be eligible for dispatching in
                # another dispatching session.
                # If we need to begin to address the remaining queued items before
                # this (the queue processor) task is awaited, we could insert a
                # Condition here to alert some sort of previously scheduled special
                # tear-down task.
                raise MissingImplementationError(f"Executor has no implementation for {str(command)}.")

            key = command["add_item"]
            edit_item = executor.get_edit_item()
            with edit_item(key) as item:
                if not isinstance(item, Task):
                    raise InternalError(f"Bug: Expected {edit_item} to return a _context.Task")

                # TODO: Check task dependencies.
                ##
                # Note that we could insert resource management here. We could create
                # tasks until we run out of free resources, then switch modes to awaiting
                # tasks until resources become available, then switch back to placing
                # tasks.

                # Bind the WorkflowManager item to an RP Task.
                task = await updater.submit(item=item)

                if task.done():
                    # Stop processing the queue if task was cancelled or errored.
                    # TODO: Test this logical branch or consider removing it.
                    if task.cancelled():
                        logger.info(f"Stopping queue processing after unexpected cancellation of task {task}")
                        return
                    else:
                        exc = task.exception()
                        if exc:
                            logger.error(f"Task {task} failed much to fast. Stopping execution.")
                            raise exc
                        else:
                            logger.debug(f"Task {task} already done. Continuing.")
                else:
                    executor.submitted_tasks.add(task)

        except Exception as e:
            logger.debug("Leaving queue runner due to exception.")
            raise e
        finally:
            logger.debug('Releasing "{}" from command queue.'.format(str(command)))
            queue.task_done()


# TODO: Resolve circular reference between `execution` and `workflow` modules.
# class ExecutorFactory(typing.Protocol[_BackendT]):
#     def __call__(self,
#                  manager: WorkflowManager,
#                  params: typing.Optional[_BackendT] = None) -> RuntimeManager[_BackendT]:
#         ...



"""Local execution dispatching for ScaleMS workflows.

Usage:
    python3 -m scalems.local my_workflow.py

"""

import sys

import scalems.invocation

if __name__ == "__main__":
    sys.exit(scalems.invocation.run(scalems.local.workflow_manager))



"""Python invocation of SCALE-MS workflow scripts.

Refer to :doc:`invocation` for user-level documentation about SCALE-MS command line
invocation.

The base command line parser is provided by :py:func:`scalems.invocation.base_parser`,
extended (optionally) by the :ref:`backend`, and further extended by
:py:func:`scalems.invocation.run`. Get usage for a particular backend with
reference to the particular module.

Workflow Manager
----------------

Managed workflows are dispatched to custom execution back-ends through
:py:func:`run`, which accepts a WorkflowManager creation function as its argument.
Most of the customization hooks are provided through the implementing module.
I.e. the *manager_factory* argument has its ``__module__`` attribute queried
to get the implementing *module*.

.. py:currentmodule:: <module>

Required Attributes
~~~~~~~~~~~~~~~~~~~
Execution back-end modules (modules providing a *manager_factory*) *MUST* provide
the following module attribute(s).

.. py:attribute:: parser
    :noindex:

    `argparse.ArgumentParser` for the execution module. For correct composition,
    see `scalems.invocation.make_parser()`.

Optional Attributes
~~~~~~~~~~~~~~~~~~~
Execution back-end modules (modules providing a *manager_factory*) *MAY* provide
the following module attribute(s) for hooks in `run()`

.. py:attribute:: logger
    :noindex:

    `logging.Logger` instance to use for the invocation.

.. py:attribute:: configuration
    :noindex:

    A callable to initialize and retrieve the current module configuration.
    If present in *module*, `run()` calls ``module.configuration(args)``,
    where *args* is a :py:class:`argparse.Namespace` object created by the
    module's *parser*.

"""
import argparse
import asyncio
import functools
import os
import runpy
import sys
import threading
import typing

import scalems.exceptions
import scalems.workflow
from scalems import ScriptEntryPoint

_reentrance_guard = threading.Lock()


# We can import scalems.context and set module state before using runpy to
# execute the script in the current process. This allows us to preconfigure a
# default execution manager.

# TODO: Consider whether we want launched scripts to have `__name__` set to `__main__` or not.

# TODO: Consider whether we want to parse execution module arguments, including handling chained `-m`.
#     Consider generalizing this boilerplate.

# TODO: Support REPL (e.g. https://github.com/python/cpython/blob/3.8/Lib/asyncio/__main__.py)


def run_dispatch(work, context: scalems.workflow.WorkflowManager):
    """Run the provided work in the execution dispatching context.

    Parameters:
        context (scalems.workflow.WorkflowManager) : an active workflow manager (with a running event loop)
        work (typing.Callable) : An "app" function to run within the scope of an active execution dispatcher.

    Initially, we assume that *work* is a `scalems.app` decorated function, though
    this is not technically required in order to call this function directly.
    Semantics and typing will presumably be clarified and tightened in the future.
    """

    async def _dispatch(_work):
        async with context.dispatch():
            # Add work to the queue
            _work()
        # Return an iterable of results.
        # for task in context.tasks: ...
        ...

    _loop = context.loop()
    _coro = _dispatch(work)
    # TODO: Name the task, if `work` has suitable attribute(s).
    _task = _loop.create_task(_coro)
    _result = _loop.run_until_complete(_task)
    return _result


class _ManagerT(typing.Protocol):
    """Primary argument type for scalems.invocation.run().

    An imported callable object with the signature defined by this Protocol.

    We use a typing.Protocol instead of typing.Callable to emphasize that the object of
    this type is an object with the full interface described for User-defined functions in
    the Python Data Model for Callable types.
    See https://docs.python.org/3/reference/datamodel.html#the-standard-type-hierarchy.
    """

    def __call__(self, loop: asyncio.AbstractEventLoop) -> scalems.workflow.WorkflowManager:
        ...


def run(manager_factory: _ManagerT, _loop: asyncio.AbstractEventLoop = None):  # noqa: C901
    """Execute boilerplate for scalems entry point scripts.

    Provides consistent logic for command line invocation of a script with a main `scalems.app` function.

    Supports invocation of the following form with minimal ``backend/__main__.py``

        python -m scalems.radical --venv=/path/to/venv --resource=local.localhost myscript.py arg1 --foo bar

    See `scalems.invocation` module documentation for details about the expected *manager_factory* module
    interface.

    Refer to module documentation for the execution backend for command line arguments.
    Unrecognized command line arguments will be passed along to the called script through
    modification to `sys.argv`.
    """
    safe = _reentrance_guard.acquire(blocking=False)
    if not safe:
        raise RuntimeError("scalems launcher is not reentrant.")
    try:
        module = sys.modules[manager_factory.__module__]

        parser = getattr(module, "parser", None)
        if parser is None:
            raise scalems.exceptions.APIError("Execution manager modules must provide a `parser` module member.")

        # Strip the current __main__ file from argv. Collect arguments for this script
        # and for the called script.
        args, script_args = parser.parse_known_args(sys.argv[1:])

        if args.pycharm:
            try:
                # noinspection PyUnresolvedReferences
                import pydevd_pycharm

                pydevd_pycharm.settrace("host.docker.internal", port=12345, stdoutToServer=True, stderrToServer=True)
            except ImportError:
                ...

        if not os.path.exists(args.script):
            # TODO: Support REPL (e.g. https://github.com/python/cpython/blob/3.8/Lib/asyncio/__main__.py)
            raise RuntimeError("Need a script to execute.")

        level = args.log_level
        if level is not None:
            import logging

            character_stream = logging.StreamHandler()
            # Optional: Set log level.
            logging.getLogger("scalems").setLevel(level)
            character_stream.setLevel(level)
            # Optional: create formatter and add to character stream handler
            formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
            character_stream.setFormatter(formatter)
            # add handler to logger
            logging.getLogger("scalems").addHandler(character_stream)
        logger = getattr(module, "logger")

        configure_module = getattr(module, "configuration", None)
        if configure_module is not None:
            config = configure_module(args)
            logger.debug(f"Configuration: {config}")

        sys.argv = [args.script] + script_args

        if _loop is not None:
            # Allow event loop to be provided (for debugging and testing purposes).
            loop = _loop
        else:
            # Start the asyncio event loop on behalf of the client.
            # We want to do this exactly once per invocation, and we do not want the scalems
            # package module or any particular scalems object to own the event loop.
            loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        # TODO: Clarify event loop management scheme.
        #     Do we want scripts to be like "apps" that get called with asyncio.run(),
        #     should we effectively reimplement asyncio.run through scalems.run, or
        #     should we think about
        #     [ast.PyCF_ALLOW_TOP_LEVEL_AWAIT](https://docs.python.org/3/whatsnew/3.8.html#builtins)

        # Execute the script in the current process.
        # TODO: Can we support mixing invocation with pytest?
        exitcode = 0

        try:
            with scalems.workflow.scope(manager_factory(loop)) as manager:
                try:
                    globals_namespace = runpy.run_path(args.script)

                    main = None
                    for name, ref in globals_namespace.items():
                        if isinstance(ref, ScriptEntryPoint):
                            if main is not None:
                                raise scalems.exceptions.DispatchError(
                                    "Multiple apps in the same script is not (yet?) supported."
                                )
                            main = ref
                            ref.name = name
                    if main is None:
                        raise scalems.exceptions.DispatchError("No scalems.app callables found in script.")
                    # The scalems.run call hierarchy goes through utility.run
                    # to utility._run to AsyncWorkflowManager.run,
                    # which then goes through WorkflowManager.dispatch().
                    # We should
                    # (a) clean this up,
                    # (b) return something sensible,
                    # (c) clarify error behavior.
                    # We probably don't want to be calling scalems.run() from the entry point script.
                    # `scalems.run()` and `python -m scalems.backend script.py` are alternative
                    # ways to separate the user from the `with Manager.dispatch():` block.
                    # For the moment, we can disable the `scalems.run()` mechanism, I think.
                    # cmd = scalems.run(main, context=context)

                    logger.debug("Starting asyncio run()")
                    try:
                        run_dispatch(main, manager)
                    except Exception as e:
                        logger.exception("Unhandled exception in scalems runner calling dispatch(): " + str(e))
                        raise e
                    finally:
                        loop.close()
                    assert loop.is_closed()

                    logger.debug("Finished asyncio run()")
                except SystemExit as e:
                    exitcode = e.code
        except Exception as e:
            print(f"{module} encountered Exception: {repr(e)}")
            if exitcode == 0:
                exitcode = 1
        if exitcode != 0:
            raise SystemExit(exitcode)
    finally:
        _reentrance_guard.release()


@functools.cache
def base_parser(add_help=False):
    """Get the base scalems argument parser.

    Provides a base argument parser for scripts or other module parsers.

    By default, the returned ArgumentParser is created with ``add_help=False``
    to avoid conflicts when used as a *parent* for a parser more local to the caller.
    If *add_help* is provided, it is passed along to the ArgumentParser created
    in this function.

    See Also:
         https://docs.python.org/3/library/argparse.html#parents
    """
    from . import __version__ as _scalems_version

    _parser = argparse.ArgumentParser(add_help=add_help)

    _parser.add_argument("--version", action="version", version=f"scalems version {_scalems_version}")

    _parser.add_argument(
        "--log-level",
        type=str.upper,
        choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"],
        help="Optionally configure console logging to the indicated level.",
    )

    _parser.add_argument(
        "--pycharm",
        action="store_true",
        default=False,
        help="Attempt to connect to PyCharm remote debugging system, where appropriate.",
    )

    _parser.add_argument(
        "script",
        metavar="script-to-run.py",
        type=str,
        help="The workflow script. Must contain a function decorated with `scalems.app`",
    )

    return _parser


def make_parser(module: str, parents: typing.Iterable[argparse.ArgumentParser] = None):
    """Make a SCALE-MS Execution Module command line argument parser.

    Args:
        module: Name of the execution module.
        parents: Optional list of parent parsers.

    If *parents* is not provided, `scalems.invocation.base_parser` is used to generate a default.
    If *parents* _is_ provided, one of the provided parents **should** inherit from
    `scalems.invocation.base_parser` using the *parents* parameter of
    :py:class:`argparse.ArgumentParser`.

    Notes:
        :py:attr:`__package__` module attribute and :py:attr:`__module__` class or
        function attribute are convenient for programmatically finding the *module* argument.
    """
    if parents is None:
        parents = [base_parser()]
    _parser = argparse.ArgumentParser(
        prog=module,
        description=f"Command line interface for `{module}` workflow execution module.",
        usage=f"python -m {module} <{module} args> script-to-run.py.py " "<script args>",
        parents=parents,
    )
    return _parser



"""Python logging facilities use the built-in logging module.

Upon import, the scalems package sets a placeholder "NullHandler" to block
propagation of log messages to the `handler of last resort
<https://docs.python.org/3/howto/logging.html#what-happens-if-no-configuration-is-provided>`__
(and to `sys.stderr`).

If you want to see logging output on `sys.stderr`, attach a
`logging.StreamHandler` to the 'scalems' logger.

Example::

    character_stream = logging.StreamHandler()
    # Optional: Set log level.
    logging.getLogger('scalems').setLevel(logging.DEBUG)
    character_stream.setLevel(logging.DEBUG)
    # Optional: create formatter and add to character stream handler
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    character_stream.setFormatter(formatter)
    # add handler to logger
    logging.getLogger('scalems').addHandler(character_stream)

To handle log messages that are issued while importing :py:mod:`scalems` and its submodules,
attach the handler before importing :py:mod:`scalems`. Note that if :py:mod:`scalems.radical`
will be used, you should import :py:mod:`radical.pilot` before importing `logging` to avoid spurious warnings.

Refer to submodule documentation for hierarchical loggers to allow
granular control of log handling (e.g. ``logging.getLogger('scalems.radical')``).
Refer to the Python :py:mod:`logging` module for information on connecting to and handling
logger output.
"""

__all__ = ["logger"]

# Import system facilities
from logging import DEBUG
from logging import getLogger
from logging import NullHandler

# Define `logger` attribute that is used by submodules to create sub-loggers.
logger = getLogger("scalems")
# By default, prevent scalems logs from propagating to the root logger (and to sys.stderr)
# if the user does not take action to handle logging.
logger.addHandler(NullHandler(level=DEBUG))



"""Manage the RADICAL Pilot start-up and shut-down.

The provided Runtime class encapsulates stateful resources that, once acquired,
should be shut down explicitly. Runtime instances may be used as context managers to
ensure the proper protocol is followed, or the caller can take responsibility for
calling Runtime.close() to shut down.

Note: Consider whether Runtime context manager is reentrant or multi-use.

The Runtime state encapsulates several nested states. A Session, TaskManager,
and PilotManager must be created for the Runtime to be usable. Additionally, Pilots and
scheduler tasks may be added or removed during the Runtime lifetime. To better support
the alternate scenarios when a Runtime instance may be provided to a scalems.radical
component in an arbitrary state, consider making the ``with`` block scoped, such that
it only returns the Runtime instance to its initial state when exiting, rather than
shutting down completely.
See also https://docs.python.org/3/library/contextlib.html#contextlib.ExitStack

Deferred:
    Runtime can avoid providing direct access to RP interface, and instead run an
    entire RP Session state machine in a thread (separate from the asyncio event loop
    thread), relaying RP scripting commands through queues, in order to completely
    prevent misuse and to insulate the asyncio event loop from blocking RP commands.
    We need to get a better sense of the RP flow combinatorics before we can reasonably
    pursue this.

See Also:
    https://github.com/SCALE-MS/scale-ms/issues/55

.. uml::

    title scalems on radical.pilot run time

    box "SCALE-MS framework" #honeydew
    participant WorkflowManager as client_workflowmanager
    participant RuntimeManager
    participant "runner task" as scalems.execution
    end box
    box "SCALE-MS RP adapter" #linen
    participant Runtime as client_runtime
    participant Executor as client_executor
    end box

    autoactivate on

    client_workflowmanager -> client_executor: async with executor
    activate client_workflowmanager
    client_executor -> RuntimeManager: ~__aenter__()
    RuntimeManager -> client_executor: runtime_configuration()
    RuntimeManager -> client_executor: runtime_startup()

    client_executor -> : rp.Session()
    return
    client_executor -> client_runtime **: Session
    activate client_runtime
    client_executor -> : rp.PilotManager()
    return
    client_executor -> client_runtime: pilot_manager()
    return
    client_executor -> : rp.TaskManager()
    return
    client_executor -> client_runtime: task_manager()
    return
    client_executor -> : pilot_manager.submit_pilots()
    return
    client_executor -> client_runtime: pilot()
    note left
    Pilot venv is determined by resource definition (JSON file).
    end note
    return

    group ref [scalems.radical.raptor]
    client_executor -> client_executor: get_scheduler()
    return
    end

    client_executor -> client_runtime: set scheduler
    return

    client_executor -> scalems.execution **: create_task(manage_execution)
    client_executor -> scalems.execution: await runner_started
    RuntimeManager <-- client_executor: set runner_task
    deactivate client_executor
    deactivate RuntimeManager

    client_workflowmanager -> client_workflowmanager #gray: async with dispatcher

    ...Raptor workload handling...

    return leave dispatcher context

    == Shut down runtime ==

    client_executor -> RuntimeManager: ~__aexit__()
    RuntimeManager -> RuntimeManager: await runner_task
    RuntimeManager <-- scalems.execution
    deactivate RuntimeManager
    RuntimeManager -> client_executor: runtime_shutdown()

    client_runtime <- client_executor
    return session
    client_runtime <- client_executor
    return scheduler
    group Cancel Master task [if scheduler is not None]
    client_runtime <- client_executor
    return task_manager
    client_executor -> : task_manager.cancel_tasks()
    return
    client_runtime <- client_executor
    return scheduler
    client_executor -> : runtime.scheduler.wait()
    return
    end
    client_executor -> : session.close()
    return

    client_executor -> client_runtime !!
    deactivate client_executor
    deactivate RuntimeManager

    client_workflowmanager <-- client_executor: leave executor context
    deactivate client_workflowmanager

"""

from __future__ import annotations

__all__ = (
    "configuration",
    "executor_factory",
    "parser",
    "Configuration",
    "Runtime",
)

import argparse
import asyncio
import collections.abc
import contextlib
import contextvars
import copy
import dataclasses
import functools
import io
import json
import logging
import os
import pathlib
import shutil
import tempfile
import threading
import typing
import uuid
import warnings
import weakref
import zipfile
from typing import Awaitable

import packaging.version
import radical.saga
import radical.utils
from radical import pilot as rp

import scalems.call
import scalems.exceptions
import scalems.execution
import scalems.file
import scalems.invocation
import scalems.messages
import scalems.radical
import scalems.radical.raptor
import scalems.store
import scalems.subprocess
import scalems.workflow
from scalems.exceptions import APIError
from scalems.exceptions import DispatchError
from scalems.exceptions import InternalError
from scalems.exceptions import MissingImplementationError
from scalems.exceptions import ProtocolError
from scalems.exceptions import ScaleMSError
from .raptor import master_input
from .raptor import master_script
from ..store import FileStore
from ..execution import AbstractWorkflowUpdater
from ..execution import RuntimeManager
from ..identifiers import EphemeralIdentifier
from ..identifiers import TypeIdentifier

logger = logging.getLogger(__name__)
logger.debug("Importing {}".format(__name__))

# TODO: Consider scoping for these context variables.
# Need to review PEP-567 and PEP-568 to consider where and how to scope the Context
# with respect to the dispatching scope.
_configuration = contextvars.ContextVar("_configuration")


if typing.TYPE_CHECKING:
    import radical.utils as ru


def _parse_option(arg: str) -> tuple:
    if not isinstance(arg, str):
        raise InternalError("Bug: This function should only be called with a str.")
    if arg.count("=") != 1:
        raise argparse.ArgumentTypeError('Expected a key/value pair delimited by "=".')
    return tuple(arg.split("="))


class RPConfigurationError(ScaleMSError):
    """Unusable RADICAL Pilot configuration."""


@functools.cache
def parser(add_help=False):
    """Get the module-specific argument parser.

    Provides a base argument parser for scripts using the scalems.radical backend.

    By default, the returned ArgumentParser is created with ``add_help=False``
    to avoid conflicts when used as a *parent* for a parser more local to the caller.
    If *add_help* is provided, it is passed along to the ArgumentParser created
    in this function.

    See Also:
         https://docs.python.org/3/library/argparse.html#parents
    """
    _parser = argparse.ArgumentParser(add_help=add_help, parents=[scalems.invocation.base_parser()])

    # We could consider inferring a default venv from the VIRTUAL_ENV environment
    # variable,
    # but we currently have very poor error handling regarding venvs. For now, this needs
    # to be explicit.
    # Ref https://github.com/SCALE-MS/scale-ms/issues/89
    # See also https://github.com/SCALE-MS/scale-ms/issues/90
    # TODO: Set module variables rather than carry around an args namespace?

    _parser.add_argument(
        "--access",
        type=str,
        help="Explicitly specify the access_schema to use from the RADICAL resource.",
    )

    _parser.add_argument(
        "--enable-raptor",
        action="store_true",
        help="Enable RP Raptor, and manage an execution side dispatching task.",
    )

    _parser.add_argument(
        "--pilot-option",
        action="append",
        type=_parse_option,
        metavar="<key>=<value>",
        help="Add a key value pair to the `radical.pilot.PilotDescription`.",
    )

    _parser.add_argument(
        "--resource",
        type=str,
        required=True,
        help="Specify a `RP resource` for the radical.pilot.PilotDescription. (Required)",
    )

    _parser.add_argument(
        "--venv",
        metavar="PATH",
        type=str,
        required=True,
        help="Path to the (pre-configured) Python virtual "
        "environment with which RP tasks should be executed. "
        "(Required. See also https://github.com/SCALE-MS/scale-ms/issues/90)",
    )

    return _parser


@dataclasses.dataclass(frozen=True)
class Configuration:
    """Module configuration information.

    See also:
        * :py:func:`scalems.radical.configuration`
        * :py:data:`scalems.radical.runtime.parser`
        * :py:class:`scalems.radical.runtime.Runtime`

    .. todo:: Consider merging with module Runtime state container.

    """

    # Note that the use cases for this dataclass interact with module ContextVars,
    # pending refinement.
    datastore: FileStore = None
    execution_target: str = "local.localhost"
    rp_resource_params: dict = dataclasses.field(default_factory=dict)
    target_venv: str = None
    enable_raptor: bool = False

    def __post_init__(self):
        hpc_platform_label = self.execution_target
        access = self.rp_resource_params["PilotDescription"].get("access_schema")
        try:
            job_endpoint: ru.Url = rp.utils.misc.get_resource_job_url(hpc_platform_label, schema=access)
        except (TypeError, KeyError) as e:
            raise RPConfigurationError(f"Could not resolve {access} access for {hpc_platform_label}") from e

        if self.enable_raptor:
            # scalems uses the RP MPIWorker, which can have problems in "local" execution modes.
            launch_scheme = job_endpoint.scheme
            if launch_scheme == "fork":
                message = f"RP Raptor MPI Worker not supported for '{launch_scheme}' launch method."
                message += f" '{access}' access for {hpc_platform_label}: {job_endpoint}"
                raise RPConfigurationError(message)


class Runtime:
    """Container for scalems.radical runtime state data.

    Note:
        This class is almost exclusively a container. Lifetime management is assumed
        to be handled externally.

    .. todo:: Let `scalems.radical.runtime.Configuration` holds the run time invariant.
        And let this become more than a container, explicitly encapsulating
        the responsibilities of `RPDispatchingExecutor.runtime_startup()`
        and `RPDispatchingExecutor.runtime_shutdown()`.

    See Also:
        :py:attr:`scalems.radical.runtime.RPDispatchingExecutor.runtime`

    """

    _session: rp.Session

    scheduler: typing.Optional[rp.Task] = None
    """The active raptor scheduler task, if any."""

    resources: typing.Optional[asyncio.Task[dict]] = None
    """The active Pilot resources, if any.

    The runtime_startup routine schedules a Task to get a copy of
    the Pilot.resource_details['rm_info'] dictionary, once the Pilot
    reaches state PMGR_ACTIVE.
    """

    _pilot_manager: typing.Optional[rp.PilotManager] = None
    _pilot: typing.Optional[rp.Pilot] = None
    _task_manager: typing.Optional[rp.TaskManager] = None

    def __init__(self, session: rp.Session):
        if not isinstance(session, rp.Session) or session.closed:
            raise ValueError("*session* must be an active RADICAL Pilot Session.")
        self._session = session

    def __repr__(self):
        if session := self._session:
            session = session.uid
        if pilot := self._pilot:
            pilot = pilot.uid
        if scheduler := self.scheduler:
            scheduler = scheduler.uid
        representation = f'<Runtime session:"{session}" pilot:"{pilot}" raptor:"{scheduler}">'
        return representation

    def reset(self, session: rp.Session):
        """Reset the runtime state.

        Close any existing resources and revert to a new Runtime state containing only
        the provided *session*.
        """
        if not isinstance(session, rp.Session) or session.closed:
            raise ValueError("*session* must be an active RADICAL Pilot Session.")
        if self.resources is not None:
            self.resources.cancel()
            self.resources = None
        self._session.close()
        # Warning: This is not quite right.
        # The attribute values are deferred to the class dict from initialization. The
        # following lines actually leave the instance members in place with None values
        # rather than removing them, but the logic of checking for and removing the
        # instance values seems a little harder to read.
        # TODO: Properly close previous session. This includes sending a "stop" to the
        #     raptor scheduler instead of letting it get a `Task.cancel()` (i.e. a SIGTERM).
        self.scheduler = None
        self._pilot = None
        self._task_manager = None
        self._pilot_manager = None
        self._session = session

    @property
    def session(self) -> rp.Session:
        """The current radical.pilot.Session (may already be closed)."""
        return self._session

    @typing.overload
    def pilot_manager(self) -> typing.Union[rp.PilotManager, None]:
        ...

    @typing.overload
    def pilot_manager(self, pilot_manager: str) -> rp.PilotManager:
        ...

    @typing.overload
    def pilot_manager(self, pilot_manager: rp.PilotManager) -> rp.PilotManager:
        ...

    def pilot_manager(self, pilot_manager=None) -> typing.Union[rp.PilotManager, None]:
        """Get (optionally set) the current PilotManager.

        Args:
            pilot_manager (optional, radical.pilot.PilotManager, str):
                Set to RP PilotManager instance or identifier, if provided.

        Returns:
            radical.pilot.PilotManager: instance, if set, else ``None``.

        Raises:
            ValueError: for invalid identifier.
            APIError: for invalid RP Session configuration.
        """
        if pilot_manager is None:
            return self._pilot_manager
        elif isinstance(pilot_manager, rp.PilotManager):
            if not pilot_manager.session.uid == self.session.uid:
                raise APIError("Cannot accept a PilotManager from a different Session.")
            self._pilot_manager = pilot_manager
            return pilot_manager
        else:
            uid = pilot_manager
            try:
                pmgr = self.session.get_pilot_managers(pmgr_uids=uid)
                assert isinstance(pmgr, rp.PilotManager)
            except (AssertionError, KeyError) as e:
                raise ValueError(f"{uid} does not describe a valid PilotManager") from e
            except Exception as e:
                logger.exception("Unhandled RADICAL Pilot exception.", exc_info=e)
                raise ValueError(f"{uid} does not describe a valid PilotManager") from e
            else:
                return self.pilot_manager(pmgr)

    @typing.overload
    def task_manager(self) -> typing.Union[rp.TaskManager, None]:
        ...

    @typing.overload
    def task_manager(self, task_manager: str) -> rp.TaskManager:
        ...

    @typing.overload
    def task_manager(self, task_manager: rp.TaskManager) -> rp.TaskManager:
        ...

    def task_manager(self, task_manager=None) -> typing.Union[rp.TaskManager, None]:
        """Get (optionally set) the current TaskManager.

        Args:
            task_manager (optional, radical.pilot.TaskManager, str):
                Set to RP TaskManager instance or identifier, if provided.

        Returns:
            radical.pilot.TaskManager: instance, if set, else ``None``.

        Raises:
            ValueError: for invalid identifier.
            APIError: for invalid RP Session configuration.
        """
        if task_manager is None:
            return self._task_manager
        elif isinstance(task_manager, rp.TaskManager):
            if not task_manager.session.uid == self.session.uid:
                raise APIError("Cannot accept a TaskManager from a different Session.")
            self._task_manager = task_manager
            return task_manager
        else:
            uid = task_manager
            try:
                tmgr = self.session.get_task_managers(tmgr_uids=uid)
                assert isinstance(tmgr, rp.TaskManager)
            except (AssertionError, KeyError) as e:
                raise ValueError(f"{uid} does not describe a valid TaskManager") from e
            except Exception as e:
                logger.exception("Unhandled RADICAL Pilot exception.", exc_info=e)
                raise ValueError(f"{uid} does not describe a valid TaskManager") from e
            else:
                return self.task_manager(tmgr)

    @typing.overload
    def pilot(self) -> typing.Union[rp.Pilot, None]:
        ...

    @typing.overload
    def pilot(self, pilot: str) -> rp.Pilot:
        ...

    @typing.overload
    def pilot(self, pilot: rp.Pilot) -> rp.Pilot:
        ...

    def pilot(self, pilot=None) -> typing.Union[rp.Pilot, None]:
        """Get (optionally set) the current Pilot.

        Args:
            pilot (radical.pilot.Pilot, str, None): Set to RP Pilot instance or identifier, if provided.

        Returns:
            radical.pilot.Pilot: instance, if set, else ``None``

        Raises:
            ValueError: for invalid identifier.
            APIError: for invalid RP Session configuration.
        """
        if pilot is None:
            return self._pilot

        pmgr = self.pilot_manager()
        if not pmgr:
            raise APIError("Cannot set Pilot before setting PilotManager.")

        if isinstance(pilot, rp.Pilot):
            session = pilot.session
            if not isinstance(session, rp.Session):
                raise APIError(f"Pilot {repr(pilot)} does not have a valid Session.")
            if session.uid != self.session.uid:
                raise APIError("Cannot accept a Pilot from a different Session.")
            if pilot.pmgr.uid != pmgr.uid:
                raise APIError("Pilot must be associated with a PilotManager already configured.")
            self._pilot = pilot
            return pilot
        else:
            uid = pilot
            try:
                pilot = pmgr.get_pilots(uids=uid)
                assert isinstance(pilot, rp.Pilot)
            except (AssertionError, KeyError, ValueError) as e:
                raise ValueError(f"{uid} does not describe a valid Pilot") from e
            except Exception as e:
                # TODO: Track down the expected rp exception.
                logger.exception("Unhandled RADICAL Pilot exception.", exc_info=e)
                raise ValueError(f"{uid} does not describe a valid Pilot") from e
            else:
                return self.pilot(pilot)


async def _get_scheduler(
    pre_exec: typing.Iterable[str],
    task_manager: rp.TaskManager,
    filestore: FileStore,
    scalems_env: str,
):
    """Establish the radical.pilot.raptor.Master task.

    Create a master rp.Task (running the scalems_rp_master script) with the
    provided *name* to be referenced as the *scheduler* for raptor tasks.

    Returns the rp.Task for the master script once the Master is ready to
    receive submissions.

    Raises:
        DispatchError if the master task could not be launched successfully.

    Note:
        Currently there is no completion condition for the master script.
        Caller is responsible for canceling the Task returned by this function.
    """
    # define a raptor.scalems master and launch it within the pilot
    td = rp.TaskDescription()

    # The master uid is used as the `scheduler` value for raptor task routing.
    # TODO(#108): Use caller-provided *name* for master_identity.
    # Master tasks may not appear unique, but must be uniquely identified within the
    # scope of a rp.Session for RP bookkeeping. Since there is no other interesting
    # information at this time, we can generate a random ID and track it in our metadata.
    master_identity = EphemeralIdentifier()
    td.uid = "scalems-rp-master." + str(master_identity)

    td.mode = rp.RAPTOR_MASTER

    # scalems_rp_master will write output before it begins handling requests. The
    # script may crash even before it can write anything, but if it does write
    # anything, we _will_ have the output file locally.
    # TODO: Why don't we have the output files? Can we ensure output files for CANCEL?
    td.output_staging = []  # TODO(#229) Write and stage output from master task.
    td.stage_on_error = True

    td.pre_exec = list(pre_exec)

    # We are not using prepare_env at this point. We use the `venv` configured by the
    # caller.
    # td.named_env = 'scalems_env'

    # This is the name that should be resolvable in an active venv for the installed
    # master task entry point script
    td.executable = master_script()

    logger.debug(f"Using {filestore}.")

    # _original_callback_duration = asyncio.get_running_loop().slow_callback_duration
    # asyncio.get_running_loop().slow_callback_duration = 0.5
    config_file = await asyncio.create_task(
        master_input(filestore=filestore, worker_pre_exec=list(pre_exec), worker_venv=scalems_env),
        name="get-master-input",
    )
    # asyncio.get_running_loop().slow_callback_duration = _original_callback_duration

    # TODO(#75): Automate handling of file staging directives for scalems.file.AbstractFileReference
    # e.g. _add_file_dependency(td, config_file)
    config_file_name = str(td.uid) + "-config.json"
    td.input_staging = [
        {
            "source": config_file.as_uri(),
            "target": f"{config_file_name}",
            "action": rp.TRANSFER,
        }
    ]
    td.arguments.append(config_file_name)

    task_metadata = {"uid": td.uid, "task_manager": task_manager.uid}

    await asyncio.create_task(asyncio.to_thread(filestore.add_task, master_identity, **task_metadata), name="add-task")
    # filestore.add_task(master_identity, **task_metadata)

    logger.debug(f"Launching RP raptor scheduling. Submitting {td}.")

    _task = asyncio.create_task(asyncio.to_thread(task_manager.submit_tasks, td), name="submit-Master")
    scheduler: rp.Task = await _task

    # WARNING: rp.Task.wait() *state* parameter does not handle tuples, but does not
    # check type.
    _task = asyncio.create_task(
        asyncio.to_thread(scheduler.wait, state=[rp.states.AGENT_EXECUTING] + rp.FINAL),
        name="check-Master-started",
    )
    await _task
    logger.debug(f"Scheduler in state {scheduler.state}.")
    # TODO: Generalize the exit status checker for the Master task and perform this
    #  this check at the call site.
    if scheduler.state in rp.FINAL:
        if scheduler.stdout or scheduler.stderr:
            logger.error(f"scheduler.stdout: {scheduler.stdout}")
            logger.error(f"scheduler.stderr: {scheduler.stderr}")
        raise DispatchError(f"Master Task unexpectedly reached {scheduler.state} during launch.")
    return scheduler


# functools can't cache this function while Configuration is unhashable (due to
# unhashable dict member).
# @functools.cache
def get_pre_exec(conf: Configuration) -> tuple:
    """Get the sequence of pre_exec commands for tasks on the currently configured execution target.

    Warning:
        Use cases may require a `list` object. Caller is responsible for converting
        the returned tuple if appropriate.

    """
    if conf.target_venv is None or len(conf.target_venv) == 0:
        raise ValueError("Currently, tasks cannot be dispatched without a target venv.")

    activate_venv = ". " + str(os.path.join(conf.target_venv, "bin", "activate"))
    # Note: RP may specifically expect a `list` and not a `tuple`.
    sequence = (activate_venv,)
    return sequence


@functools.singledispatch
def _set_configuration(*args, **kwargs) -> Configuration:
    """Initialize or retrieve the module configuration.

    This module and the RADICAL infrastructure have various stateful aspects
    that require clearly-scoped module-level configuration. Module configuration
    should be initialized exactly once per Python process.

    Recommended usage is to derive an ArgumentParser from the *parser()* module
    function and use the resulting namespace to initialize the module configuration
    using this function.

    Note this is a dispatch function. "Overloads" are defined in separate decorated
    functions for calls that provide a `Configuration` or `argparse.Namespace`
    object as the first (and only) positional argument.
    """
    assert len(args) != 0 or len(kwargs) != 0
    # Caller has provided arguments.
    # Not thread-safe
    if _configuration.get(None):
        raise APIError(f"configuration() cannot accept arguments when {__name__} is already configured.")
    c = Configuration(*args, **kwargs)
    _configuration.set(c)
    return _configuration.get()


@_set_configuration.register
def _(config: Configuration) -> Configuration:
    # Not thread-safe
    if _configuration.get(None):
        raise APIError(f"configuration() cannot accept arguments when {__name__} is already configured.")
    _configuration.set(config)
    return _configuration.get()


@_set_configuration.register
def _(namespace: argparse.Namespace) -> Configuration:
    rp_resource_params = {
        "PilotDescription": {
            "access_schema": namespace.access,
            "exit_on_error": False,
        }
    }
    if namespace.pilot_option is not None and len(namespace.pilot_option) > 0:
        user_options = _PilotDescriptionProxy.normalize_values(namespace.pilot_option)
        rp_resource_params["PilotDescription"].update(user_options)
        logger.debug(f'Pilot options: {repr(rp_resource_params["PilotDescription"])}')

    if namespace.enable_raptor:
        logger.debug("RP Raptor enabled.")
    else:
        logger.debug("RP Raptor disabled.")

    config = Configuration(
        execution_target=namespace.resource,
        target_venv=namespace.venv,
        rp_resource_params=rp_resource_params,
        enable_raptor=namespace.enable_raptor,
    )
    return _set_configuration(config)


async def new_session():
    """Start a new RADICAL Pilot Session.

    Returns:
        Runtime instance.

    """
    # Note that we cannot resolve the full _resource config until we have a Session
    # object.
    # We cannot get the default session config until after creating the Session,
    # so we don't have a template for allowed, required, or default values.
    # Question: does the provided *cfg* need to be complete? Or will it be merged
    # with default values from some internal definition, such as by dict.update()?
    # I don't remember what the use cases are for overriding the default session
    # config.
    session_config = None

    # Note: We may soon want Session ID to be deterministic (or to be re-used?).
    session_id = None

    # Note: radical.pilot.Session creation causes several deprecation warnings.
    # Ref https://github.com/radical-cybertools/radical.pilot/issues/2185
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=DeprecationWarning)
        # This would be a good time to `await`, if an event-loop friendly
        # Session creation function becomes available.
        session_args = dict(uid=session_id, cfg=session_config)
        _task = asyncio.create_task(asyncio.to_thread(rp.Session, (), **session_args), name="create-Session")
        session = await _task
        runtime = Runtime(session=session)
    return runtime


@functools.singledispatch
def normalize(hint: object, value):
    """Dispatching value normalizer.

    Normalize value according to hint.

    Raises:
        MissingImplementationError: if key could not be dispatched for normalization.
        TypeError: if value could not be normalized according to hint.

    """
    raise MissingImplementationError(f"No dispatcher for {repr(value)} -> {repr(hint)}.")


@normalize.register
def _(hint: type, value):
    return hint(value)


@normalize.register
def _(hint: list, value):
    if len(hint) != 1:
        raise InternalError(f"Expected a list of one type element. Got {repr(hint)}.")
    if isinstance(value, (str, bytes)) or not isinstance(value, collections.abc.Iterable):
        raise TypeError(f"Expected a list-like value. Got {repr(value)}.")
    return [normalize(hint[0], element) for element in value]


@normalize.register
def _(hint: dict, value):
    try:
        for key in value.keys():
            if key not in hint:
                raise MissingImplementationError(f"{key} is not a valid field.")
        items: tuple = value.items()
    except AttributeError:
        raise TypeError(f"Expected a dict-like value. Got {repr(value)}.")
    return {key: normalize(hint[key], val) for key, val in items}


class _PilotDescriptionProxy(rp.PilotDescription):
    """Use PilotDescription details to normalize the value types of description fields."""

    assert hasattr(rp.PilotDescription, "_schema")
    assert isinstance(rp.PilotDescription._schema, dict)
    assert all(
        map(
            lambda v: isinstance(v, (type, list, dict, type(None))),
            rp.PilotDescription._schema.values(),
        )
    )

    @classmethod
    def normalize_values(cls, desc: typing.Sequence[tuple]):
        """Generate normalized key-value tuples.

        For values that are not already of the appropriate type, cast according to
        PilotDescription._schema.

        Args:
            desc: sequence of key-value tuples for PilotDescription fields.

        Raises:
            MissingImplementationError: if key could not be dispatched for normalization.
            TypeError: if value could not be normalized according to hint.

        """
        for key, value in desc:
            try:
                hint = cls._schema[key]
            except KeyError:
                raise MissingImplementationError(f"{key} is not a valid PilotDescription field.")
            if not isinstance(hint, type):
                # This may be overly aggressive, but at the moment we are only normalizing values from
                # the command line parser, and we don't have a good way to pre-parse list or dict values.
                raise MissingImplementationError(f"No handler for {key} field of type {repr(hint)}.")

            if isinstance(None, hint) or isinstance(value, hint):
                yield key, value
            else:
                yield key, normalize(hint, value)


async def add_pilot(runtime: Runtime, **kwargs):
    """Get a new rp.Pilot and add it to the Runtime instance."""
    if kwargs:
        message = (
            f"scalems.radical.runtime.add_pilot() version {scalems.__version__} "
            + "does not support keyword arguments "
            + ", ".join(kwargs.keys())
        )
        raise MissingImplementationError(message)
    pilot_manager = runtime.pilot_manager()
    task_manager = runtime.task_manager()
    if not pilot_manager or not task_manager:
        raise ProtocolError(f"Runtime {runtime} is not ready to use.")

    # Note: Consider fetching through the Runtime instance.
    config: Configuration = configuration()

    # Warning: The Pilot ID needs to be unique within the Session.
    pilot_description = {"uid": f"pilot.{str(uuid.uuid4())}"}
    pilot_description.update(config.rp_resource_params.get("PilotDescription", {}))
    pilot_description.update({"resource": config.execution_target})

    # TODO: Pilot venv (#90, #94).
    # Currently, Pilot venv must be specified in the JSON file for resource
    # definitions.
    pilot_description = rp.PilotDescription(pilot_description)
    logger.debug("Submitting PilotDescription {}".format(repr(pilot_description.as_dict())))
    pilot: rp.Pilot = await asyncio.create_task(
        asyncio.to_thread(pilot_manager.submit_pilots, pilot_description), name="submit_pilots"
    )
    logger.debug(f"Got Pilot {pilot.uid}: {pilot.as_dict()}")
    runtime.pilot(pilot)

    await asyncio.create_task(asyncio.to_thread(task_manager.add_pilots, pilot), name="add_pilots")

    return pilot


async def get_pilot_resources(pilot: rp.Pilot):
    def log_pilot_state(fut: asyncio.Task[str]):
        # Task callbacks do not get called for cancelled Tasks.
        assert not fut.cancelled()
        if e := fut.exception():
            logger.exception("Exception while watching for Pilot to become active.", exc_info=e)
        logger.info(f"Pilot {pilot.uid} in state {pilot.state}.")

    logger.info("Waiting for an active Pilot.")
    # Wait for Pilot to be in state PMGR_ACTIVE. (There is no reasonable
    # choice of a timeout because we are waiting for the HPC queuing system.)
    # Then, query Pilot.resource_details['rm_info']['requested_cores'] and 'requested_gpus'.
    pilot_state = asyncio.create_task(
        asyncio.to_thread(pilot.wait, state=rp.PMGR_ACTIVE, timeout=None), name="pilot_state_waiter"
    )

    pilot_state.add_done_callback(log_pilot_state)
    await pilot_state
    rm_info: dict = pilot.resource_details.get("rm_info")
    logger.debug(f"Pilot {pilot.uid} resources: {str(rm_info)}")
    if rm_info is not None:
        assert isinstance(rm_info, dict)
        assert "requested_cores" in rm_info and isinstance(rm_info["requested_cores"], int)
        assert "requested_gpus" in rm_info and isinstance(rm_info["requested_gpus"], typing.SupportsFloat)
        return rm_info.copy()


class RPDispatchingExecutor(RuntimeManager):
    """Client side manager for work dispatched through RADICAL Pilot.

    Configuration points:

    * resource config
    * pilot config
    * session config?

    Extends :py:class:`scalems.execution.RuntimeManager`
    """

    runtime: "scalems.radical.runtime.Runtime"
    """See `scalems.execution.RuntimeManager.runtime`"""

    def __init__(
        self,
        *,
        editor_factory: typing.Callable[[], typing.Callable] = None,
        datastore: FileStore = None,
        loop: asyncio.AbstractEventLoop,
        configuration: Configuration,
        dispatcher_lock=None,
    ):
        """Create a client side execution manager.

        Warning:
            The creation method does not fully initialize the instance.

            Initialization and de-initialization occurs through
            the Python (async) context manager protocol.

        """
        if "RADICAL_PILOT_DBURL" not in os.environ:
            raise DispatchError("RADICAL Pilot environment is not available.")

        if not isinstance(configuration.target_venv, str) or len(configuration.target_venv) == 0:
            raise ValueError("Caller must specify a venv to be activated by the execution agent for dispatched tasks.")

        super().__init__(
            editor_factory=editor_factory,
            datastore=datastore,
            loop=loop,
            configuration=configuration,
            dispatcher_lock=dispatcher_lock,
        )

    @contextlib.contextmanager
    def runtime_configuration(self):
        """Provide scoped Configuration.

        Merge the runtime manager's configuration with the global configuration,
        update the global configuration, and yield the configuration for a ``with`` block.

        Restores the previous global configuration when exiting the ``with`` block.

        Warning:
            Not thread-safe.

            We do not check for re-entrance, which will cause race conditions w.r.t.
            which Context state is restored! Moreover, the Configuration object is not
            currently hashable and does not have an equality test defined.

            This contextmanager is not async, but it could be (and is) used within
            an asynchronous context manager, so we don't have anything structurally
            prohibiting reentrant calls, even without multithreading.

        TODO:
            Reconsider logic for runtime module scoped Configuration management.

            Instead of managing the current contextvars.Context,
            this should be used to produce an updated contextvars.Context for use by
            a Task or Context.run() scope. Alternatively, we can decouple the runtime configuration
            from the global module configuration, reduce the dynamic utility of
            *<module>.configuration()*, and not use contextvars.

            Design notes:
                Do we want two-way interaction between module
                and instance configuration? Under what circumstances will one or the other
                change during execution? Should we be providing the configuration through
                the current Context, through a Context instance (usable for Context.run() or
                to the task launching command), or simply as a Configuration object?
        """

        # Get default configuration.
        configuration_dict = dataclasses.asdict(configuration())
        # Update with any internal configuration.
        if self._runtime_configuration.target_venv is not None and len(self._runtime_configuration.target_venv) > 0:
            configuration_dict["target_venv"] = self._runtime_configuration.target_venv
        if len(self._runtime_configuration.rp_resource_params) > 0:
            configuration_dict["rp_resource_params"].update(self._runtime_configuration.rp_resource_params)
        if (
            self._runtime_configuration.execution_target is not None
            and len(self._runtime_configuration.execution_target) > 0
        ):
            configuration_dict["execution_target"] = self._runtime_configuration.execution_target
        configuration_dict["datastore"] = self.datastore

        c = Configuration(**configuration_dict)

        token = _configuration.set(c)
        try:
            yield c
        finally:
            _configuration.reset(token)

    async def runtime_startup(self) -> asyncio.Task:
        """Establish the RP Session.

        Acquire a maximally re-usable set of RP resources. The scope established by
        this function is as broad as it can be within the life of the workflow manager.

        Once *instance.runtime_startup()* succeeds, *instance.runtime_shutdown()*
        must be called to clean up resources.
        Use the async context manager behavior of the instance to
        automatically follow this protocol. I.e. instead of calling
        ``instance.runtime_startup(); ...; instance.runtime_shutdown()``,
        use::

            async with instance:
                ...

        Raises:
            DispatchError: if task dispatching could not be set up.
            asyncio.CancelledError: if parent `asyncio.Task` is cancelled while executing.

        TODO: More concurrency.
            The rp.Pilot and raptor task can be separately awaited, and we should allow
            input data staging to begin as soon as we have enough run time details to do it.
            We need to clarify which tasks should be in which state to consider the
            asynchronous context manager to have been successfully "entered". I expect
            that success includes a Pilot in state PMGR_ACTIVE, and a raptor Task in
            state AGENT_EXECUTING, and asyncio.Task handles available for other aspects,
            like synchronization of metadata and initiation of input data staging.
            However, we may prefer that the workflow script can continue evaluation on
            the client side while waiting for the Pilot job, and that we avoid blocking
            until we absolutely have to (presumably when exiting the dispatching context).

        """
        config: Configuration = configuration()

        # TODO: Check that we have a FileStore.

        # We try to wrap rp UI calls in separate threads. Note, though, that
        #   * The rp.Session needs to be created in the root thread to be able to correctly
        #     manage signal handlers and subprocesses, and
        #   * We need to be able to schedule RP Task callbacks in the same process as the
        #     asyncio event loop in order to handle Futures for RP tasks.
        # See https://github.com/SCALE-MS/randowtal/issues/2

        # TODO: RP triggers SIGINT in various failure modes.
        #  We should use loop.add_signal_handler() to convert to an exception
        #  that we can raise in an appropriate task. However, we should make sure that we
        #  account for the signal handling that RP expects to be able to do.
        #  See https://github.com/SCALE-MS/randowtal/issues/1
        #  We should also make sure that scalems and RP together handle keyboard interrupts
        #  as a user would expect.
        # Note that PilotDescription can use `'exit_on_error': False` to suppress the SIGINT,
        # but we have not fully explored the consequences of doing so.
        # Note also that RP Task.cancel() issues a SIGTERM in Task processes,
        # which we would normally be even more cautious about trapping, so we should
        # avoid explicitly or implicitly allowing Task.cancel().

        try:
            #
            # Start the Session.
            #

            # Note: the current implementation implies that only one Task for the dispatcher
            # will exist at a time. We are further assuming that there will probably only
            # be one Task per the lifetime of the dispatcher object.
            # We could choose another approach and change our assumptions, if appropriate.
            logger.debug("Entering RP dispatching context. Waiting for rp.Session.")

            _runtime: Runtime = await new_session()
            # At some point soon, we need to track Session ID for the workflow metadata.
            session_id = _runtime.session.uid
            # Do we want to log this somewhere?
            # session_config = copy.deepcopy(self.session.cfg.as_dict())
            logger.debug("RP dispatcher acquired session {}".format(session_id))

            logger.debug("Launching PilotManager.")
            pilot_manager = await asyncio.create_task(
                asyncio.to_thread(rp.PilotManager, session=_runtime.session),
                name="get-PilotManager",
            )
            logger.debug("Got PilotManager {}.".format(pilot_manager.uid))
            _runtime.pilot_manager(pilot_manager)

            logger.debug("Launching TaskManager.")
            task_manager = await asyncio.create_task(
                asyncio.to_thread(rp.TaskManager, session=_runtime.session),
                name="get-TaskManager",
            )
            logger.debug(("Got TaskManager {}".format(task_manager.uid)))
            _runtime.task_manager(task_manager)

            #
            # Get a Pilot
            #
            # We can launch an initial Pilot, but we may have to run further Pilots
            # during self._queue_runner_task (or while servicing scalems.wait() within the
            # with block) to handle dynamic work load requirements.
            # Optionally, we could refrain from launching the pilot here, at all,
            # but it seems like a good chance to start bootstrapping the agent environment.
            #
            # How and when should we update the pilot description?

            pilot = await add_pilot(runtime=_runtime)
            logger.debug("Added Pilot {} to task manager {}.".format(pilot.uid, _runtime.task_manager().uid))

            # TODO: Asynchronous data staging optimization.
            #  Allow input data staging to begin before scheduler is in state EXECUTING and
            #  before Pilot is in state PMGR_ACTIVE.

            # Note: This could take hours or days depending on the queuing system.
            # Can we report some more useful information, like job ID?
            _runtime.resources = asyncio.create_task(get_pilot_resources(pilot))

            assert _runtime.scheduler is None

            # Get a scheduler task IFF raptor is explicitly enabled.
            if config.enable_raptor:
                # Note that _get_scheduler is a coroutine that, itself, returns a rp.Task.
                # We await the result of _get_scheduler, then store the scheduler Task.
                _runtime.scheduler = await asyncio.create_task(
                    _get_scheduler(
                        pre_exec=list(get_pre_exec(config)),
                        task_manager=task_manager,
                        filestore=config.datastore,
                        scalems_env="scalems_venv",
                        # TODO: normalize ownership of this name.
                    ),
                    name="get-scheduler",
                )  # Note that we can derive scheduler_name from self.scheduler.uid in later methods.
        except asyncio.CancelledError as e:
            raise e
        except Exception as e:
            logger.exception("Exception while connecting RADICAL Pilot.", exc_info=e)
            raise DispatchError("Failed to launch SCALE-MS master task.") from e

        self.runtime = _runtime

        if self.runtime is None or self.runtime.session.closed:
            raise ProtocolError("Cannot process queue without a RP Session.")

        # Launch queue processor (proxy executor).
        # TODO: Make runtime_startup optional. Let it return a resource that is
        #  provided to the normalized run_executor(), or maybe use it to configure the
        #  Submitter that will be provided to the run_executor.
        runner_started = asyncio.Event()
        runner_task = asyncio.create_task(scalems.execution.manage_execution(self, processing_state=runner_started))
        await runner_started.wait()
        # TODO: Note the expected scope of the runner_task lifetime with respect to
        #  the global state changes (i.e. ContextVars and locks).
        return runner_task

    @staticmethod
    async def cpi(command: str, runtime: Runtime):
        """Send a control command to the raptor scheduler.

        Implements :py:func:`scalems.execution.RuntimeManager.cpi()`

        TODO: Unify with new raptor cpi feature.
        """
        timeout = 180

        logger.debug(f'Received command "{command}" for runtime {runtime}.')
        if runtime is None:
            raise scalems.exceptions.ScopeError("Cannot issue control commands without an active RuntimeManager.")
        scheduler: rp.Task = runtime.scheduler
        logger.debug(f"Preparing command for {repr(scheduler)}.")
        if scheduler is None:
            raise scalems.exceptions.ScopeError(
                f"Cannot issue control commands without an active RuntimeManager. {runtime}"
            )
        if scheduler.state in rp.FINAL:
            raise scalems.exceptions.ScopeError(f"Raptor scheduler is not available. {repr(scheduler)}")
        assert scheduler.uid
        message = scalems.messages.Control.create(command)
        td = rp.TaskDescription(
            from_dict={
                "scheduler": scheduler.uid,
                "mode": scalems.radical.raptor.CPI_MESSAGE,
                "metadata": message.encode(),
                "uid": EphemeralIdentifier(),
            }
        )
        logger.debug(f"Submitting {str(td.as_dict())}")
        (task,) = await asyncio.to_thread(runtime.task_manager().submit_tasks, [td])
        logger.debug(f"Submitted {str(task.as_dict())}. Waiting...")
        # Warning: we can't wait on the final state of such an rp.Task unless we
        # _also_ watch the scheduler itself, because there are various circumstances
        # in which the Task may never reach a rp.FINAL state.

        command_watcher = asyncio.create_task(
            asyncio.to_thread(task.wait, state=rp.FINAL, timeout=timeout), name="cpi-watcher"
        )

        scheduler_watcher = asyncio.create_task(
            asyncio.to_thread(scheduler.wait, state=rp.FINAL, timeout=timeout), name="scheduler-watcher"
        )
        # If master task fails, command-watcher will never complete.
        done, pending = await asyncio.wait(
            (command_watcher, scheduler_watcher), timeout=timeout, return_when=asyncio.FIRST_COMPLETED
        )
        if scheduler_watcher in done and command_watcher in pending:
            command_watcher.cancel()
        logger.debug(str(task.as_dict()))
        # WARNING: Dropping the Task reference will cause its deletion.
        return command_watcher

    @staticmethod
    def runtime_shutdown(runtime: Runtime):
        """Manage tear down of the RADICAL Pilot Session and resources.

        Several aspects of the RP runtime interface use blocking calls.
        This method should be run in a non-root thread (concurrent.futures.Future)
        that the event loop can manage as an asyncio-wrapped task.

        Overrides :py:class:`scalems.execution.RuntimeManager`
        """
        session: rp.Session = getattr(runtime, "session", None)
        if session is None:
            raise scalems.exceptions.APIError(f"No Session in {runtime}.")
        if session.closed:
            logger.error("Runtime Session is already closed?!")
        else:
            if runtime.scheduler is not None:
                # Note: The __aexit__ for the RuntimeManager makes sure that a `stop`
                # is issued after the work queue is drained, if the scheduler task has
                # not already ended. We could check the status of this stop message...
                if runtime.scheduler.state not in rp.FINAL:
                    logger.info(f"Waiting for RP Raptor master task {runtime.scheduler.uid} to complete...")
                    runtime.scheduler.wait(rp.FINAL, timeout=60)
                if runtime.scheduler.state not in rp.FINAL:
                    # Cancel the master.
                    logger.warning("Canceling the master scheduling task.")
                    # Note: the effect of CANCEL is to send SIGTERM to the shell that
                    # launched the task process.
                    # TODO: Report incomplete tasks and handle the consequences of terminating the
                    #  Raptor processes early.
                    task_manager = runtime.task_manager()
                    task_manager.cancel_tasks(uids=runtime.scheduler.uid)
                # As of https://github.com/radical-cybertools/radical.pilot/pull/2702,
                # we do not expect `cancel` to block, so we must wait for the
                # cancellation to succeed. It shouldn't take long, but it is not
                # instantaneous or synchronous. We hope that a minute is enough.
                final_state = runtime.scheduler.wait(state=rp.FINAL, timeout=10)
                logger.debug(f"Final state: {final_state}")
                logger.info(f"Master scheduling task state {runtime.scheduler.state}: {repr(runtime.scheduler)}.")
                if runtime.scheduler.stdout:
                    # TODO(#229): Fetch actual stdout file.
                    logger.debug(runtime.scheduler.stdout)
                if runtime.scheduler.stderr:
                    # TODO(#229): Fetch actual stderr file.
                    # Note that all of the logging output goes to stderr, it seems,
                    # so we shouldn't necessarily consider it an error level event.
                    logger.debug(runtime.scheduler.stderr)  # TODO(#108,#229): Receive report of work handled by Master.

            # Note: there are no documented exceptions or errors to check for,
            # programmatically. Some issues encountered during shutdown will be
            # reported through the reporter or logger of the
            # radical.pilot.utils.component.Component base.
            # The RP convention seems to be to use the component uid as the name
            # of the underlying logging.Logger node, so we could presumably attach
            # a log handler to the logger for a component of interest.
            logger.debug(f"Closing Session {session.uid}.")
            session.close(download=True)

            if session.closed:
                logger.debug(f"Session {session.uid} closed.")
            else:
                logger.error(f"Session {session.uid} not closed!")
        logger.debug("Runtime shut down.")

    def updater(self) -> "WorkflowUpdater":
        return WorkflowUpdater(executor=self)


def configuration(*args, **kwargs) -> Configuration:
    """Get (and optionally set) the RADICAL runtime configuration.

    With no arguments, returns the current configuration. If a configuration has
    not yet been set, the command line parser is invoked to try to build a new
    configuration.

    If arguments are provided, try to construct a `scalems.radical.runtime.Configuration`
    and use it to initialize the module.

    It is an error to try to initialize the module more than once.
    """
    # Not thread-safe
    from scalems.radical import parser

    if len(args) > 0:
        _set_configuration(*args, **kwargs)
    elif len(kwargs) > 0:
        _set_configuration(Configuration(**kwargs))
    elif _configuration.get(None) is None:
        # No config is set yet. Generate with module parser.
        namespace, _ = parser.parse_known_args()
        _set_configuration(namespace)
    return _configuration.get()


def executor_factory(manager: scalems.workflow.WorkflowManager, params: Configuration = None):
    if params is not None:
        _set_configuration(params)
    params = configuration()

    executor = RPDispatchingExecutor(
        editor_factory=weakref.WeakMethod(manager.edit_item),
        datastore=manager.datastore(),
        loop=manager.loop(),
        configuration=params,
        dispatcher_lock=manager._dispatcher_lock,
    )
    return executor


class RPResult:
    """Basic result type for RADICAL Pilot tasks.

    Define a return type for Futures or awaitable tasks from
    RADICAL Pilot commands.
    """

    # TODO: Provide support for RP-specific versions of standard SCALEMS result types.


class RPTaskFailure(ScaleMSError):
    """Error in radical.pilot.Task execution.

    Attributes:
        failed_task: A dictionary representation of the failed task.

    TODO: What can/should we capture and report from the failed task?
    """

    failed_task: dict

    def __init__(self, *args, task: rp.Task):
        super().__init__(*args)
        self.failed_task = task.as_dict()


class RPInternalError(ScaleMSError):
    """RADICAL Pilot is misbehaving, probably due to a bug.

    Please report the potential bug.
    """


class RPFinalTaskState:
    # TODO: Provide a bridge between the threading and asyncio Event primitives so that
    #  we can effectively `await` for the events. Presumably this would look like an
    #  asyncio.Future created from a threading.Future of a watcher of the three Events.
    #  Use asyncio.wrap_future() to wrap a threading.Future to an asyncio.Future.
    #  If there is not an adapter, we can loop.run_in_executor() to run the
    #  threading.Event watcher in a separate thread.
    # For future reference, check out the `sched` module.
    def __init__(self):
        self.canceled = threading.Event()
        self.done = threading.Event()
        self.failed = threading.Event()

    def __bool__(self):
        return self.canceled.is_set() or self.done.is_set() or self.failed.is_set()

    def __repr__(self):
        return (
            f"<{self.__class__.__qualname__} "
            f"canceled={self.canceled.is_set()} "
            f"done={self.done.is_set()} "
            f"failed={self.failed.is_set()}>"
        )


def _rp_callback(
    obj: rp.Task,
    state,
    cb_data: weakref.ReferenceType = None,
    final: RPFinalTaskState = None,
):
    """Prototype for RP.Task callback.

    To use, partially bind the *final* parameter (with `functools.partial`) to get a
    callable with the RP.Task callback signature.

    Register with *task* to be called when the rp.Task state changes.
    """
    if final is None:
        raise APIError("This function is strictly for dynamically prepared RP callbacks through " "functools.partial.")
    logger.debug(f"Callback triggered by {repr(obj)} state change to {repr(state)}.")
    try:
        # Note: assertions and exceptions are not useful in RP callbacks.
        if state in (rp.states.DONE, rp.states.CANCELED, rp.states.FAILED):
            # TODO: Pending https://github.com/radical-cybertools/radical.pilot/issues/2444
            # tmgr: rp.TaskManager = obj.tmgr
            # ref = cb_data()
            # tmgr.unregister_callback(cb=ref, metrics='TASK_STATE',
            #                          uid=obj.uid)
            logger.debug(f"Recording final state {state} for {repr(obj)}")
            # Consider alternatives that are easier to await for:
            #   * use an asyncio.Event instead, and use loop.call_soon_threadsafe(<event>.set)
            #   * insert a layer so that the RPFinalTaskState object can easily
            #     set its own awaitable "done" status when updated
            if state == rp.states.DONE:
                final.done.set()
            elif state == rp.states.CANCELED:
                final.canceled.set()
            elif state == rp.states.FAILED:
                final.failed.set()
            else:
                logger.error("Bug: logic error in state cases.")
    except Exception as e:
        logger.error(f"Exception encountered during rp.Task callback: {repr(e)}")


async def _rp_task_watcher(task: rp.Task, final: RPFinalTaskState, ready: asyncio.Event) -> rp.Task:  # noqa: C901
    """Manage the relationship between an RP.Task and a scalems Future.

    Cancel the RP.Task if this task or the scalems.Future is canceled.

    Publish the RP.Task result or cancel the scalems.Future if the RP.Task is
    done or canceled.

    Arguments:
        task: RADICAL Pilot Task, submitted by caller.
        final: thread-safe event handler for the RP task call-back to announce it has run.
        ready: output parameter, set when coroutine has run enough to perform its
            responsibilities.

    Returns:
        *task* in its final state.

    An asyncio.Future based on this coroutine has very similar semantics to the
    required *future* argument, but this is subject to change.
    The coroutine is intended to facilitate progress of the task,
    regardless of the rp.Task results. The provided *future* allows the rp.Task
    results to be interpreted and semantically translated. rp.Task failure is
    translated into an exception on *future*. The *future* has a different
    exposure than the coroutine return value, as well: again, the *future* is
    connected to the workflow item and user-facing interface, whereas this
    coroutine is a detail of the task management. Still, these two modes of output
    are subject to revision without notice.

    Caller should await the *ready* event before assuming the watcher task is doing its
    job.
    """

    async def wait_for_final(state: RPFinalTaskState) -> RPFinalTaskState:
        """Function to watch for final event.

        This is a poor design. We should replace with a queuing system for state
        transitions as described in scalems.workflow.
        """
        while not state:
            # TODO: What is the right adapter between asyncio and threading event waits?
            await asyncio.sleep(0.05)
        return state

    event_watcher = asyncio.create_task(wait_for_final(final))

    try:
        ready.set()

        while not event_watcher.done():
            # TODO(#96): Use a control thread to manage *threading* primitives and
            #  translate to asyncio primitives.
            # Let the watcher wake up periodically to check for suspicious state.

            _rp_task_was_complete = task.state in rp.FINAL

            done, pending = await asyncio.wait([event_watcher], timeout=60.0, return_when=asyncio.FIRST_COMPLETED)

            if _rp_task_was_complete and not event_watcher.done():
                event_watcher.cancel()
                raise RPInternalError(
                    f"RP Callbacks are taking too long to complete. Abandoning {repr(task)}. Please report bug."
                )

            if event_watcher in done:
                assert final
                logger.debug(f"Handling finalization for RP task {task.uid}.")
                if final.failed.is_set():
                    logger.error(f"{task.uid} stderr: {task.stderr}")
                    logger.info(f"Failed {task.uid} working directory: {task.task_sandbox}")
                    if logger.getEffectiveLevel() <= logging.DEBUG:
                        for key, value in task.as_dict().items():
                            logger.debug(f"    {key}: {str(value)}")
                    message = f"{task.uid} failed."
                    if task.exit_code is not None:
                        message += f" Exit code {task.exit_code}."
                    if task.stderr is not None:
                        message += f" stderr: {task.stderr}"
                    if task.exception is not None:
                        message += f" Exception: {task.exception}"
                    raise RPTaskFailure(message, task=task)
                elif final.canceled.is_set():
                    # Act as if RP called Task.cancel() on us.
                    raise asyncio.CancelledError()
                assert final.done.is_set()

                logger.debug(f"Publishing results from RP Task {task.uid}.")
                # TODO: Manage result type.
                return task

        raise scalems.exceptions.InternalError("Logic error. This line should not have been reached.")

    except asyncio.CancelledError as e:
        logger.debug(f"Received cancellation in watcher task for {repr(task)}")
        if task.state not in rp.CANCELED:
            logger.debug(f"Propagating cancellation to {repr(task)}.")
            task.cancel()
        raise e


# TODO: Separate this out to a scalems.rptask module.
async def rp_task(rptask: rp.Task) -> asyncio.Task[rp.Task]:
    """Mediate between a radical.pilot.Task and an asyncio.Future.

    Schedule an asyncio Task to receive the result of the RP Task. The asyncio
    Task must also make sure that asyncio cancellation propagates to the rp.Task.cancel,
    and vice versa.

    This function should be awaited immediately to make sure the necessary call-backs
    get registered. The result will be an asyncio.Task, which should be awaited
    separately.

    Internally, this function provides a call-back to the rp.Task. The call-back
    provided to RP cannot directly call asyncio.Future methods (such as set_result() or
    set_exception()) because RP will be making the call from another thread without
    mediation by the asyncio event loop.

    As such, we provide a thread-safe event handler to propagate the
    RP Task call-back to to this asyncio.Task result.
    (See :py:func:`_rp_callback()` and :py:class:`RPFinalTaskState`)

    Canceling the returned task will cause *rptask* to be canceled.
    Canceling *rptask* will cause this task to be canceled.

    Arguments:
        rptask: RADICAL Pilot Task that has already been submitted.

    Returns:
        A Task that, when awaited, returns the rp.Task instance in its final state.
    """
    if not isinstance(rptask, rp.Task):
        raise TypeError("Function requires a RADICAL Pilot Task object.")

    final = RPFinalTaskState()
    callback = functools.partial(_rp_callback, final=final)
    functools.update_wrapper(callback, _rp_callback)

    # Note: register_callback() does not provide a return value to use for
    # TaskManager.unregister_callback and we cannot provide *callback* with a reference
    # to itself until after it is created, so we will get a reference here that we can
    # provide through the *cb_data* argument of rp.Task callbacks.
    cb_data: weakref.ReferenceType = weakref.ref(callback)

    rptask.register_callback(callback, cb_data=cb_data, metric=rp.constants.TASK_STATE)

    if rptask.state in rp.FINAL:
        # rptask may have reached FINAL state before callback was registered.
        # radical.pilot.Task.register_callback does not guarantee that callbacks
        # will be called if registered after task completion.
        # Call it once. For simplicity, let the task_watcher logic proceed normally.
        logger.warning(f"RP Task {repr(rptask)} finished suspiciously fast.")
        callback(rptask, rptask.state, cb_data)

    watcher_started = asyncio.Event()
    waiter = asyncio.create_task(watcher_started.wait())
    wrapped_task = asyncio.create_task(_rp_task_watcher(task=rptask, final=final, ready=watcher_started))

    # Make sure that the task is cancellable before returning it to the caller.
    await asyncio.wait((waiter, wrapped_task), return_when=asyncio.FIRST_COMPLETED)
    if wrapped_task.done():
        # Let CancelledError propagate.
        e = wrapped_task.exception()
        if e is not None:
            raise e
    # watcher_task.
    return wrapped_task


def _describe_legacy_task(item: scalems.workflow.Task, pre_exec: list) -> rp.TaskDescription:
    """Derive a RADICAL Pilot TaskDescription from a scalems workflow item.

    For a "raptor" style task, see _describe_raptor_task()
    """
    subprocess_type = TypeIdentifier(("scalems", "subprocess", "SubprocessTask"))
    assert item.description().type() == subprocess_type
    input_data = item.input
    task_input = scalems.subprocess.SubprocessInput(**input_data)
    args = list([arg for arg in task_input.argv])
    # Warning: TaskDescription class does not have a strongly defined interface.
    # Check docs for schema.
    task_description = rp.TaskDescription(
        from_dict=dict(
            executable=args[0],
            arguments=args[1:],
            stdout=str(task_input.stdout),
            stderr=str(task_input.stderr),
            pre_exec=pre_exec,
        )
    )
    uid: str = item.uid().hex()
    task_description.uid = uid

    # TODO: Check for and activate an appropriate venv
    # using
    #     task_description.pre_exec = ...
    # or
    #     task_description.named_env = ...

    # TODO: Interpret item details and derive appropriate staging directives.
    task_description.input_staging = list(task_input.inputs.values())
    task_description.output_staging = [
        {
            "source": str(task_input.stdout),
            "target": os.path.join(uid, pathlib.Path(task_input.stdout).name),
            "action": rp.TRANSFER,
        },
        {
            "source": str(task_input.stderr),
            "target": os.path.join(uid, pathlib.Path(task_input.stderr).name),
            "action": rp.TRANSFER,
        },
    ]
    task_description.output_staging += task_input.outputs.values()

    return task_description


def _describe_raptor_task(item: scalems.workflow.Task, scheduler: str, pre_exec: list) -> rp.TaskDescription:
    """Derive a RADICAL Pilot TaskDescription from a scalems workflow item.

    The TaskDescription will be submitted to the named *scheduler*,
    where *scheduler* is the UID of a task managing the life of a rp.raptor.Master
    instance.

    Caller is responsible for ensuring that *scheduler* is valid.
    """
    # Warning: TaskDescription class does not have a strongly defined interface.
    # Check docs for schema.
    # Ref: scalems_rp_master._RaptorTaskDescription
    task_description = rp.TaskDescription(
        from_dict=dict(executable="scalems", pre_exec=pre_exec)
        # This value is currently ignored, but must be set.
    )
    task_description.uid = item.uid()
    task_description.scheduler = str(scheduler)
    # Example work would be the JSON serialized form of the following dictionary.
    # {'mode': 'call',
    #  'cores': 1,
    #  'timeout': 10,
    #  'data': {'method': 'hello',
    #           'kwargs': {'world': uid}}}
    #
    # Maybe something like this:
    # work_dict = {
    #     'mode': 'scalems',
    #     'cores': 1,
    #     'timeout': 10,
    #     'data': item.serialize()
    # }
    work_dict = {
        "mode": "exec",
        "cores": 1,
        "timeout": None,
        "data": {"exe": item.input["argv"][0], "args": item.input["argv"][1:]},
    }
    task_description.arguments = [json.dumps(work_dict)]

    # TODO: Check for and activate an appropriate venv
    # using
    #     task_description.pre_exec = ...
    # or
    #     task_description.named_env = ...

    # TODO: Interpret item details and derive appropriate staging directives.
    task_description.input_staging = []
    task_description.output_staging = []

    return task_description


async def submit(
    *,
    item: scalems.workflow.Task,
    task_manager: rp.TaskManager,
    pre_exec: list,
    scheduler: typing.Optional[str] = None,
) -> asyncio.Task:
    """Dispatch a WorkflowItem to be handled by RADICAL Pilot.

    Submits an rp.Task and returns an asyncio.Task watcher for the submitted task.

    Creates a Future, registering a done_callback to publish the task result with
    *item.set_result()*.

    A callback is registered with the rp.Task to set an Event on completion. An
    asyncio.Task watcher task monitors the Event(s) and gets a reference to an
    asyncio.Future through which results can be published to the scalems workflow item.
    (Currently the Future is created in this function, but should probably be acquired
    directly from the *item* itself.) The watcher task waits for the rp.Task
    finalization event or for the Future to be cancelled. Periodically, the watcher
    task "wakes up" to check if something has gone wrong, such as the rp.Task
    completing without setting the finalization event.

    Caveats:
        There is an unavoidable race condition in the check performed by the watcher
        task. We don't know how long after an rp.Task completes before its callbacks
        will run and we can't check whether the callback has been scheduled. The
        watcher task cannot be allowed to complete successfully until we know that the
        callback has either run or will never run.

        The delay between the rp.Task state change and the callback execution should be
        less than a second. We can allow the watcher to wake up occasionally (on the
        order of minutes), so we can assume that it will never take more than a full
        iteration of the waiting loop for the callback to propagate, unless there is a
        bug in RP. For simplicity, we can just note whether ``rptask.state in rp.FINAL``
        before the watcher goes to sleep and raise an error if the callback is not
        triggered in an iteration where we have set such a flag.

        If our watcher sends a cancellation to the rp.Task, there is no need to
        continue to monitor the rp.Task state and the watcher may exit.

    Args:
        item: The workflow item to be submitted
        task_manager: A `radical.pilot.TaskManager` instance
            through which the task should be submitted.
        pre_exec: :py:data:`radical.pilot.Task.pre_exec` prototype.
        scheduler (str): The string name of the "scheduler," corresponding to
            the UID of a Task running a rp.raptor.Master (if Raptor is enabled).

    Returns:
        asyncio.Task: a "Future[rp.Task]" for a rp.Task in its final state.

    The caller *must* await the result of the coroutine to obtain an asyncio.Task that
    can be cancelled or awaited as a proxy to direct RP task management. The Task will
    hold a coroutine that is guaranteed to already be running, failed, or canceled. The
    caller should check the status of the task immediately before making assumptions
    about whether a Future has been successfully bound to the managed workflow item.

    The returned asyncio.Task can be used to cancel the rp.Task (and the Future)
    or to await the RP.Task cleanup.

    To submit tasks as a batch, await an array of submit() results in the
    same dispatching context. (TBD)
    """

    # TODO: Optimization: skip tasks that are already done (cached results available).
    def scheduler_is_ready(scheduler):
        return (
            isinstance(scheduler, str) and len(scheduler) > 0 and isinstance(task_manager.get_tasks(scheduler), rp.Task)
        )

    subprocess_type = TypeIdentifier(("scalems", "subprocess", "SubprocessTask"))
    submitted_type = item.description().type()
    if submitted_type == subprocess_type:
        if scheduler is not None:
            raise DispatchError("Raptor not yet supported for scalems.executable.")
        rp_task_description = _describe_legacy_task(item, pre_exec=pre_exec)
    elif configuration().enable_raptor:
        if scheduler_is_ready(scheduler):
            # We might want a contextvars.Context to hold the current rp.Master instance name.
            rp_task_description = _describe_raptor_task(item, scheduler, pre_exec=pre_exec)
        else:
            raise APIError("Caller must provide the UID of a submitted *scheduler* task.")
    else:
        raise APIError(f"Cannot dispatch {submitted_type}.")

    # TODO(#249): A utility function to move slow blocking RP calls to a separate thread.
    #  Compartmentalize TaskDescription -> rp_task_watcher in a separate utility function.
    task = task_manager.submit_tasks(rp_task_description)

    rp_task_watcher = await rp_task(rptask=task)

    if rp_task_watcher.done():
        if rp_task_watcher.cancelled():
            raise DispatchError(f"Task for {item} was unexpectedly canceled during dispatching.")
        e = rp_task_watcher.exception()
        if e is not None:
            raise DispatchError("Task for {item} failed during dispatching.") from e

    # Warning: in the long run, we should not extend the life of the reference returned
    # by edit_item, and we need to consider the robust way to publish item results.
    # TODO: Translate RP result to item result type.
    rp_task_watcher.add_done_callback(functools.partial(scalems_callback, item=item))
    # TODO: If *item* acquires a `cancel` method, we need to subscribe to it and
    #  respond by unregistering the callback and canceling the future.

    return rp_task_watcher


def scalems_callback(fut: asyncio.Future, *, item: scalems.workflow.Task):
    """Propagate the results of a Future to the subscribed *item*.

    Partially bind *item* to use this as the argument to *fut.add_done_callback()*.

    Warning:
        In the long run, we should not extend the life of the reference returned
        by edit_item, and we need to consider the robust way to publish item results.

    Note:
        This is not currently RP-specific and we should look at how best to factor
        results publishing for workflow items. It may be that this function is the
        appropriate place to convert RP Task results to scalems results.
    """
    assert fut.done()
    if fut.cancelled():
        logger.info(f"Task supporting {item} has been cancelled.")
    else:
        # The following should not throw because we just checked for `done()` and
        # `cancelled()`
        e = fut.exception()
        if e:
            logger.info(f"Task supporting {item} failed: {e}")
        else:
            # TODO: Construct an appropriate scalems Result from the rp Task.
            item.set_result(fut.result())


class WorkflowUpdater(AbstractWorkflowUpdater):
    def __init__(self, executor: RPDispatchingExecutor):
        self.executor = executor
        self.task_manager = executor.runtime.task_manager()
        # TODO: Make sure we are clear about the scope of the configuration and the
        #  life time of the workflow updater / submitter.
        self._pre_exec = list(get_pre_exec(executor.configuration()))

    async def submit(self, *, item: scalems.workflow.Task) -> asyncio.Task:
        # TODO: Ensemble handling
        item_shape = item.description().shape()
        if len(item_shape) != 1 or item_shape[0] != 1:
            raise MissingImplementationError("Executor cannot handle multidimensional tasks yet.")

        task: asyncio.Task[rp.Task] = await submit(item=item, task_manager=self.task_manager, pre_exec=self._pre_exec)
        return task


@dataclasses.dataclass
class RPTaskResult:
    """A collection of data and managed objects associated with a completed RP Task.

    A rp.Task is associated with additional artifacts that are not directly tied
    to the rp.Task object, such as an arbitrary number of files that are not generally
    knowable a priori.

    Design note:
        This data structure is intentionally ignorant of wrapped Python functions.
        We should try to reconcile our multiple notions of RP task management as soon
        as possible. We should be careful to distinguish classes for managing traditional
        *executable* RP tasks from Python tasks managed with raptor functionality.
    """

    uid: str
    """:py:attr:`radical.pilot.Task.uid` identifier."""

    task_dict: dict
    """Dictionary representation from :py:func:`radical.pilot.Task.as_dict()`."""

    exit_code: int
    """Exit code of the executable task."""

    final_state: str
    """Final state of the `radical.pilot.Task`."""

    directory: radical.utils.Url
    """Resource location information for the Task working directory.

    Constructed from the filesystem_endpoint. See :doc:`design/localization`.
    """

    directory_archive: Awaitable[scalems.store.FileReference]
    """An archive of the task working directory.

    This result is separately awaitable to allow elision of unnecessary transfers.
    The exact implementation is not strongly specified; a Future usually implies
    that the result will eventually be produced, whereas a more general "awaitable"
    may not and may never be scheduled.

    A future optimization should allow individual files to be selected and remotely
    extracted, but this would require more progress on the Scale-MS structured data typing model.

    TODO: In the next iteration, use a non-local FileReference or (TBD) DirectoryReference.
    """


async def subprocess_to_rp_task(
    call_handle: scalems.call._Subprocess, dispatcher: RPDispatchingExecutor
) -> RPTaskResult:
    """Dispatch a subprocess task through the `scalems.radical` execution backend.

    Get a Future for a RPTaskResult (a collection representing a completed rp.Task).

    Schedule a RP Task and wrap with asyncio. Subscribe a dependent asyncio.Task
    that can provide the intended result type and return it as the Future.
    Schedule a call-back to clean up temporary files.

    Args:
        call_handle (scalems.call._Subprocess): result of a ``await`` on a
            :py:func:`scalems.call.function_call_to_subprocess()`
        dispatcher (RPDispatchingExecutor): An active execution manager.

    TODO:
        This logic should be integrated into a :py:func:`WorkflowManager.submit()`
        stack (currently :py:func:`WorkflowManager.add_item()`)
        and `manage_execution` loop. We really should have special handling for
        wrapped function calls as distinguished from command line executables
        through "overloads" (dispatching) of the task submission.
        Also, we should not be exposing the *dispatcher* to the user.
    """
    subprocess_dict = dict(
        stage_on_error=True,
        uid=call_handle.uid,
        executable=call_handle.executable,
        arguments=list(call_handle.arguments),
        pre_exec=list(get_pre_exec(dispatcher.configuration())),
        mode=rp.TASK_EXECUTABLE,
    )
    if configuration().enable_raptor:
        subprocess_dict["scheduler"] = dispatcher.runtime.scheduler.uid
    # Capturing stdout/stderr is a potentially unusual or unexpected behavior for
    # a Python function runner, and may collide with user assumptions or native
    # behaviors of third party tools. We will specify distinctive names for the RP
    # output capture files in the hope of clarifying the component responsible for
    # these files.
    subprocess_dict["stdout"] = "_scalems_stdout.txt"
    subprocess_dict["stderr"] = "_scalems_stderr.txt"

    # Note: We could save wall time in the job by pre-staging files, but we are
    # deferring that for future optimization.
    # Ref: https://github.com/SCALE-MS/scale-ms/issues/316
    await asyncio.gather(*tuple(ref.localize() for ref in call_handle.input_filenames.values()))
    subprocess_dict["input_staging"] = [
        {
            "source": ref.as_uri(),
            # TODO: Find a programmatic mechanism to translate between URI and CLI arg for robustness.
            "target": f"task:///{name}",
            "action": rp.TRANSFER,
        }
        for name, ref in call_handle.input_filenames.items()
    ]

    # We just pass the user-provided requirements through, but we have to reject
    # any that collide with parameters we expect to generate.
    for param, value in call_handle.requirements.items():
        if param in subprocess_dict:
            raise ValueError(f"Cannot overwrite {param}. Task['{param}'] is {subprocess_dict[param]}")
        else:
            subprocess_dict[param] = value

    try:
        subprocess_task_description = rp.TaskDescription(from_dict=subprocess_dict).verify()
    except radical.utils.typeddict.TDKeyError as e:
        raise ValueError("Invalid attribute for RP TaskDescription.") from e

    task_cores = subprocess_task_description.cores_per_rank * subprocess_task_description.cpu_processes
    rm_info = await dispatcher.runtime.resources
    pilot_cores = rm_info["requested_cores"]
    # TODO: Account for Worker cores.
    if configuration().enable_raptor:
        raptor_task: rp.Task = dispatcher.runtime.scheduler
        raptor_cores = raptor_task.description["ranks"] * raptor_task.description["cores_per_rank"]
    else:
        raptor_cores = 0
    if task_cores > (available_cores := pilot_cores - raptor_cores):
        raise ValueError(f"Requested {task_cores} for {call_handle.uid}, but at most {available_cores} are available.")

    # TODO: Find a better way to express these three lines.
    # This seems like it should be a subscription by the local workflow context to the
    # RP dispatching workflow context. Something like
    #     supporting_task = await rp_manager.add_task(task_description)
    #     supporting_task.add_done_callback(...)
    task_manager = dispatcher.runtime.task_manager()
    (submitted_task,) = await asyncio.to_thread(task_manager.submit_tasks, [subprocess_task_description])
    subprocess_task_future: asyncio.Task[rp.Task] = await scalems.radical.runtime.rp_task(submitted_task)

    # TODO: We really should consider putting timeouts on all tasks.
    subprocess_task: rp.Task = await subprocess_task_future

    logger.debug(f"Task {subprocess_task.uid} sandbox: {subprocess_task.task_sandbox}.")

    sandbox: radical.saga.Url = copy.deepcopy(subprocess_task.task_sandbox)
    if not isinstance(sandbox, radical.saga.Url):
        logger.debug(f"Converting {repr(sandbox)} to expected Url type.")
        sandbox = radical.saga.Url(sandbox)

    # TODO: Schedule the task to create the remote archive, separating the archive retrieval.
    #   Get a Future for the saga URL (`done` when the archive exists remotely).
    #   In a later implementation, get a Future[FileReference] that is non-local.
    #   A refactoring of this wrapper can immediately schedule retrieval to support
    #   the packaged result.
    # Note: We need to find language to distinguish a Future that is scheduled from
    #   an awaitable that isn't. The native Coroutine is not sufficient because Coroutines
    #   are expected to be scheduled eventually and to be awaited at some point.
    #   Can we allow all potential outputs to be scheduled, but just at low priority somehow?
    #   Or can this be a formalization of our notion of "localize" as an awaitable that
    #   changes the state of the parent object?
    archive_future = asyncio.create_task(
        get_directory_archive(sandbox, dispatcher=dispatcher),
        name=f"archive-{subprocess_task.uid}",
    )
    # In the future, we can reduce latency by separating the transfers for requested results
    # from transfers for the full archive. (Use scalems.call._Subprocess.output_filenames.)
    result = RPTaskResult(
        uid=subprocess_task.uid,
        exit_code=subprocess_task.exit_code,
        task_dict=subprocess_task.description,
        directory=sandbox,
        final_state=subprocess_task.state,
        directory_archive=archive_future,
    )
    return result


async def wrapped_function_result_from_rp_task(
    subprocess: scalems.call._Subprocess, rp_task_result: RPTaskResult
) -> scalems.call.CallResult:
    """

    Once `subprocess_to_rp_task` has produced a `RPTaskResult`
    for a `scalems.call._Subprocess`,
    we produce a `scalems.call.CallResult` with the help of some file transfers.

    Args:
        subprocess: the result of a `scalems.call.function_call_to_subprocess`
        rp_task_result: the result of a `subprocess_to_rp_task` for the *subprocess*

    Returns:
        localized result from :py:func:`scalems.call.function_call_to_subprocess()`

    """
    if rp_task_result.final_state != rp.DONE:
        # Note: it is possible that the executable task could fail for reasons other than a Python Exception.
        logger.info(f"Task for {subprocess.uid} in final state {rp_task_result.final_state}.")
        logger.debug(repr(rp_task_result))
    # It looks like this should be encapsulated with the workflow management details
    # of the dispatched work item or further specialized for this specific task type.
    # It doesn't make sense that we should need both the _Subprocess and
    # RPTaskResult objects to be passed by the caller.
    # Also, we don't have access to the requested *outputs* in order to optimize out transfers
    # without adding yet another argument.
    archive_ref = await rp_task_result.directory_archive
    # TODO: Collaborate with scalems.call to agree on output filename.
    output_file = subprocess.output_filenames[0]

    with zipfile.ZipFile(pathlib.Path(archive_ref)) as myzip:
        if logger.getEffectiveLevel() <= logging.DEBUG:
            logger.debug(f"Opening ZipFile {myzip.filename}: {', '.join(myzip.namelist())}")
        with myzip.open(output_file) as fh:
            partial_result: scalems.call.CallResult = scalems.call.deserialize_result(io.TextIOWrapper(fh).read())
    if partial_result.exception is not None:
        assert rp_task_result.final_state != rp.DONE
        logger.error(f"Subprocess {subprocess.uid} encountered Exception: {partial_result.exception}")
    # The remotely packaged result was not aware of the task environment managed by RP.
    # Update and augment the managed file details.
    stdout = copy.deepcopy(rp_task_result.directory).set_path(
        os.path.join(rp_task_result.directory.path, rp_task_result.task_dict["stdout"])
    )
    stderr = copy.deepcopy(rp_task_result.directory).set_path(
        os.path.join(rp_task_result.directory.path, rp_task_result.task_dict["stderr"])
    )

    result = dataclasses.replace(partial_result, stdout=str(stdout), stderr=str(stderr), directory=archive_ref.as_uri())
    return result


async def get_directory_archive(
    directory: radical.saga.Url, dispatcher: RPDispatchingExecutor
) -> scalems.store.FileReference:
    """Get a local archive of a remote directory.

    TODO:
        Let this be absorbed into a collaboration between WorkflowManagement
        contexts and their local and remote data stores.
    """
    # TODO: Let FileStore tell us what directory to use.
    staging_directory = dispatcher.datastore.directory.joinpath(f"_scalems_output_staging_{id(directory)}")
    logger.debug(f"Preparing to stage {directory} to {staging_directory}.")
    assert not staging_directory.exists()

    # TODO: Don't rely on the RP Session. We should be able to do this after the
    #   Session is closed or after an interrupted Session.
    pilot = dispatcher.runtime.pilot()
    # TODO: Can we avoid serializing more than once? Such as with `rp.TARBALL`?
    await asyncio.to_thread(
        pilot.stage_out,
        sds=[
            {
                "source": directory,
                "target": staging_directory.as_uri(),
                "action": rp.TRANSFER,
                "flags": rp.CREATE_PARENTS | rp.RECURSIVE,
            }
        ],
    )
    # Work around https://github.com/radical-cybertools/radical.pilot/issues/2823
    if packaging.version.parse(rp.version) < packaging.version.parse("1.21"):
        while True:
            try:
                staging_directory.stat()
            except FileNotFoundError:
                logger.error(f"Waiting for {staging_directory} to appear...")
                await asyncio.sleep(1.0)
            else:
                break
    try:
        with tempfile.NamedTemporaryFile(mode="wb") as tmp:
            await asyncio.to_thread(
                write_archive, filehandle=tmp, root_dir=staging_directory, relative_to=staging_directory
            )
            tmp.flush()
            archive_path = pathlib.Path(tmp.name).resolve()
            file_ref = await dispatcher.datastore.add_file(scalems.file.describe_file(archive_path))
    finally:
        await asyncio.to_thread(shutil.rmtree, staging_directory)
    return file_ref


def _add_to_archive(archive: zipfile.ZipFile, source: pathlib.Path, relative_to: pathlib.Path):
    destination = source.relative_to(relative_to)
    if source.is_dir():
        for path in source.iterdir():
            _add_to_archive(archive, path, relative_to)
    else:
        if not source.is_file():
            logger.warning(
                "Directory contains unusual filesystem object. "
                f"Attempting to write {source} to {destination} in {archive.filename}"
            )
        archive.write(filename=source, arcname=destination)


def write_archive(*, filehandle, root_dir: pathlib.Path, relative_to: pathlib.Path):
    """Write a ZipFile archive.

    Args:
        filehandle: file-like object to write the archive to
        root_dir: Base of the directory tree to archive
        relative_to: Path prefix to strip from *root_dir* when constructing the paths in the archive.
    """
    assert root_dir.is_dir()
    with zipfile.ZipFile(filehandle, mode="w") as archive:
        _add_to_archive(archive, source=root_dir, relative_to=relative_to)
        if logger.getEffectiveLevel() <= logging.DEBUG:
            logger.debug(f"Wrote to {archive.filename}: {', '.join(archive.namelist())}")



"""Provide an interface for managing filesystem objects in a workflow.

We distinguish between *File* and *FileReference* types.

A *File* is a concrete representation of a local filesystem object.

A *FileReference* is more general. It may represent a non-local object
(accessible by a URI). We currently use *FileReference* objects exclusively for
resources managed by scalems. (See `scalems.store.FileStore`.)

This module has minimal dependencies on other scalems implementation details so
that it can be included easily by scalems modules without circular dependencies.
"""

__all__ = (
    "AbstractFileReference",
    "AbstractFile",
    "AccessFlags",
    "BaseBinary",
    "BaseText",
    "DataLocalizationError",
    "describe_file",
)

import abc
import codecs
import enum
import hashlib
import locale
import mmap
import os
import pathlib
import typing

import scalems.exceptions as _exceptions
from .. import identifiers as _identifiers


class DataLocalizationError(_exceptions.ScopeError):
    """The requested file operation is not possible in the given context.

    This may be because a referenced file exists in a different FileStore than the
    current default. Note that data localization is a potentially expensive operation
    and so must be explicitly requested by the user.
    """


@typing.runtime_checkable
class AbstractFileReference(typing.Protocol):
    """Reference to a managed file.

    A FileReference is an :py:class:`os.PathLike` object.

    If *filereference.is_local()* is ``True``, :py:func:`os.fspath()` will return a
    :py:class:`str` for the referenced file's path in the local filesystem.
    If the file is not available locally, an exception will be raised.

    Example:
        handle: scalems.file.AbstractFileReference = filestore.add_file(...)
    """

    @abc.abstractmethod
    def __fspath__(self) -> str:
        """Represent the local filesystem path, if possible.

        Support `os.PathLike`.

        Returns:
            str: representation of the filesystem path if the referenced file is
                available locally.

        Raises:
            DataLocalizationError: if the file has not been transferred to the local
                file store.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def is_local(self, context=None) -> bool:
        """Check for file existence in a given context.

        By default, file existence is checked in the currently active FileStore.
        If *context* is specified, checks whether the referenced file is locally
        available in the given context.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def localize(self, context=None) -> "AbstractFileReference":
        """Make sure that the referenced file is available locally for a given context.

        With no arguments, *localize()* affects the currently active FileStore.
        Specific (e.g. remote) data stores can be addressed by providing *context*.

        If the file is not yet localized, yields until the necessary data transfers
        have completed.

        Raises:
            TBD

        TODO:
            Abstract type for required *context* interface.
        """
        return self

    def as_uri(self, context=None) -> str:
        """Get an appropriate URI for a referenced file, whether or not it is available.

        For local file stores, the result is the same as for *obj.path().as_uri()* except
        that no exception is raised if the object does not yet exist locally.

        When *context* is specified, the returned URI may not be resolvable without
        additional contextual facilities. For instance, :py:mod:`scalems.radical`
        contexts may return a :py:class:`radical.saga.Url` using one of the scheme
        extensions for RADICAL Pilot
        :py:mod:`staging directives <radical.pilot.staging_directives>`.
        """
        if context is None and self.is_local():
            return pathlib.Path(self).as_uri()
        else:
            raise _exceptions.MissingImplementationError(
                "Protocol for non-local filesystem contexts is a work in progress."
            )  # return self.path(context=context).as_uri()

    @abc.abstractmethod
    def filestore(self):
        """Get a handle to the managing filestore for this reference."""
        raise NotImplementedError

    @abc.abstractmethod
    def key(self) -> _identifiers.ResourceIdentifier:
        """Get the identifying key for the file reference.

        The return value may not be human-readable, but may be necessary for locating
        the referenced file through the filestore.
        """
        raise NotImplementedError


class AccessFlags(enum.Flag):
    NO_ACCESS = 0
    READ = enum.auto()
    WRITE = enum.auto()


class AbstractFile(typing.Protocol):
    """Abstract base for file types.

    Extends `os.PathLike` with the promise of a
    :py:func:`~AbstractFile.fingerprint()` method and some support for access
    control semantics.

    See :py:func:`describe_file` for a creation function.
    """

    access: AccessFlags
    """Declared access requirements.

    Not yet enforced or widely used.
    """

    @abc.abstractmethod
    def __fspath__(self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def fingerprint(self) -> bytes:
        """Get a checksum or other suitable fingerprint for the file data.

        Note that the fingerprint could change if the file is written to during
        the lifetime of the handle. This is too complicated to optimize, generally,
        so possible caching of fingerprint results is left to the caller.
        """
        raise NotImplementedError


class BaseBinary(AbstractFile):
    """Base class for binary file types."""

    _path: str

    def __init__(self, path: typing.Union[str, pathlib.Path, os.PathLike], access: AccessFlags = AccessFlags.READ):
        self._path = str(pathlib.Path(path).resolve())
        self.access = access

    def __fspath__(self) -> str:
        return self._path

    def fingerprint(self) -> bytes:
        with open(os.fspath(self), "rb") as f:
            with mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ) as data:
                m = hashlib.sha256(data)
        checksum: bytes = m.digest()
        return checksum


class BaseText(BaseBinary):
    """Base class for text file types."""

    encoding = locale.getpreferredencoding(False)

    def __init__(self, path, access: AccessFlags = AccessFlags.READ, encoding=None):
        if encoding is None:
            encoding = locale.getpreferredencoding(False)
        try:
            codecs.getencoder(encoding)
        except LookupError:
            raise ValueError("Specify a valid character encoding for text files.")
        self.encoding = encoding
        super().__init__(path=path, access=access)

    def fingerprint(self) -> bytes:
        m = hashlib.sha256()
        # Use universal newlines and utf-8 encoding for text file fingerprinting.
        with open(self._path, "r", encoding=self.encoding) as f:
            # If we knew more about the text file size or line length, or if we already
            # had an open buffer, socket, or mmap, we could  optimize this in subclasses.
            for line in f:
                m.update(line.encode("utf8"))
        checksum: bytes = m.digest()
        return checksum


def describe_file(
    obj: typing.Union[str, pathlib.Path, os.PathLike], mode="rb", encoding: str = None, file_type_hint=None
) -> AbstractFile:
    """Describe an existing local file."""
    access = AccessFlags.NO_ACCESS
    if "r" in mode:
        access |= AccessFlags.READ
        if "w" in mode:
            access |= AccessFlags.WRITE
    else:
        # If neither 'r' nor 'w' is provided, still assume 'w'
        access |= AccessFlags.WRITE
    # TODO: Confirm requested access permissions.

    if file_type_hint:
        raise _exceptions.MissingImplementationError("Extensible file types not yet supported.")

    if "b" in mode:
        if encoding is not None:
            raise TypeError("*encoding* argument is not supported for binary files.")
        file = BaseBinary(path=obj, access=access)
    else:
        file = BaseText(path=obj, access=access, encoding=encoding)

    if not pathlib.Path(file).exists():
        raise ValueError("Not an existing file.")

    return file
